/*
 * pspawsmanger.cpp - Author: Andrew Craig
 *
 * Copyright (C) 2003 PlaneShift Team (info@planeshift.it,
 * http://www.planeshift.it)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation (version 2 of the License)
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */

// PawsManager.cpp: implementation of the PawsManager class.
//
//////////////////////////////////////////////////////////////////////
#include <cssysdef.h>

////////////////////////////////////////////////////////////////////////////
//  CS INCLUDES
////////////////////////////////////////////////////////////////////////////
#include <ivideo/graph2d.h>
#include <ivideo/graph3d.h>

#include <iutil/event.h>
#include <iutil/cfgmgr.h>
#include <csutil/event.h>

#include <csutil/objreg.h>
#include <csutil/xmltiny.h>

////////////////////////////////////////////////////////////////////////////
// COMMON INCLUDES
////////////////////////////////////////////////////////////////////////////
#include "util/log.h"

////////////////////////////////////////////////////////////////////////////
// CLIENT INCLUDES
////////////////////////////////////////////////////////////////////////////
#include "localization.h"

////////////////////////////////////////////////////////////////////////////
// PAWS INCLUDES
////////////////////////////////////////////////////////////////////////////
#include "pawsmanager.h"
#include "pawswidget.h"
#include "pawsbutton.h"
#include "pawscrollbar.h"
#include "pawsmainwidget.h"
#include "pawstextbox.h"
#include "pawstexturemanager.h"
#include "pawsmouse.h"
//#include "pawsinfowindow.h"
#include "pawsprefmanager.h"
#include "pawsobjectview.h"
#include "pawstree.h"
#include "pawsprogressbar.h"
#include "pawsmenu.h"


//////////////////////////////////////////////////////////////////////
// PAWS WIDGET FACTORIES
//////////////////////////////////////////////////////////////////////
#include "pawslistbox.h"
#include "pawsradio.h"
#include "pawsokbox.h"
#include "pawscombo.h"
#include "pawsselector.h"
#include "pawsyesnobox.h"
#include "pawscheckbox.h"
#include "pawstabwindow.h"
//#include "pawsquitinfobox.h"


//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////
PawsManager::PawsManager(iObjectRegistry* object)
{
    objectReg = object;

    graphics2D = CS_QUERY_REGISTRY( objectReg ,iGraphics2D );
    graphics3D = CS_QUERY_REGISTRY( objectReg ,iGraphics3D );

    vfs = CS_QUERY_REGISTRY( objectReg, iVFS );
    xml = csPtr<iDocumentSystem>(new csTinyDocumentSystem);

    RegisterFactories();

    mainWidget = new pawsMainWidget( this );

    textureManager = new pawsTextureManager( objectReg );

    prefs = new pawsPrefManager( this );

    if ( !prefs->LoadPrefFile( "/this/config/prefs.xml" ) )
        printf("Failed to load prefs.xml\n");
    if ( !prefs->LoadBorderFile("/this/config/borderlist.xml") )
        printf("Failed to loader borderlist.xml");

    csRef<iConfigManager> cfg = CS_QUERY_REGISTRY(objectReg, iConfigManager);
    csString imageFile = cfg->GetStr("Planeshift.GUI.ImageFile","/this/config/imagelist.xml");
                          
    if ( !textureManager->LoadImageList( imageFile ) )
        printf("Failed to load image list\n");

    mouse = new pawsMouse( this );
    mouse->ChangeImage("Standard Mouse Pointer"); 

    graphics2D->SetMouseCursor( csmcNone );

    currentFocusedWidget = mainWidget;
    movingWidget         = 0;

    resizingWidget       = 0;
    resizingFlags        = 0;
    
    modalWidget = 0;
    mouseoverWidget = 0;
	lastfadeWidget = 0;

    assert(cfg);
    csString lang = cfg->GetStr("Planeshift.Language", "");
    localization = new psLocalization();
    localization->Initialize(objectReg);
    localization->SetLanguage(lang);
    
    hadKeyDown = false;
    dragDropWidget = NULL;
}

PawsManager::~PawsManager()
{
    delete mainWidget;
    delete textureManager;
    delete prefs;
    delete localization;
	delete mouse;
    
    if (dragDropWidget != NULL)
        delete dragDropWidget;

    factories.DeleteAll();
}

void PawsManager::SetMainWidget(pawsMainWidget *widg)
{
    if (mainWidget)
        delete mainWidget;
    mainWidget = widg;
}

pawsWidget* PawsManager::FindWidget( const char* widgetName )
{
    return mainWidget->FindWidget( widgetName );
}

void PawsManager::MovingWidget( pawsWidget* widget )
{
    if ( modalWidget == NULL )
    {
        movingWidget = widget;
    }
}

void PawsManager::ResizingWidget( pawsWidget* widget, int flags )
{
    if ( modalWidget == NULL )
    {
        resizingWidget = widget;
        resizingFlags  = flags;
    }
}

bool PawsManager::HandleEvent( iEvent &event )
{
    switch ( event.Type )
    {
        case csevMouseMove: return HandleMouseMove( event );
        case csevMouseDown: return HandleMouseDown( event );
        
        case csevMouseUp  : return HandleMouseUp( event );        
    }
    
    if ( (csKeyEventHelper::GetEventType (&event) == csKeyEventTypeDown) || 
         (csKeyEventHelper::GetEventType (&event) == csKeyEventTypeUp))
        return HandleKeyDown( event );
        
        
    return false;
}

bool PawsManager::HandleKeyDown( iEvent& event )
{   
    if ( currentFocusedWidget )
    {    
        if ( (csKeyEventHelper::GetEventType (&event) == csKeyEventTypeUp) )      
        {
            if ( hadKeyDown )
                return true;
        }
         
           
        int raw = csKeyEventHelper::GetRawCode(&event);
        int cooked = csKeyEventHelper::GetCookedCode(&event);
        uint32 modifiers = csKeyEventHelper::GetModifiersBits(&event);
                        
        bool result = currentFocusedWidget->OnKeyDown( raw,        
                                                       cooked, 
                                                       modifiers );                                                       
        hadKeyDown = result;                                                       
        return result;                                                       
    }

    return false;                          
}


bool PawsManager::HandleMouseDown( iEvent &event )
{
    pawsWidget* widget = 0;
 
    widget = mainWidget->WidgetAt( event.Mouse.x, event.Mouse.y );

    if ( modalWidget != NULL )
    {
        // Check to see if the widget is a child of a the modal widget.
        // These are the only componets allowed access durning a modal 
        // mode.
        pawsWidget* check = modalWidget->FindWidget( widget->GetName() );
        if ( check != widget) 
            widget = modalWidget->WidgetAt( event.Mouse.x, event.Mouse.y );
    }

    SetCurrentFocusedWidget( widget );
       
    if ( widget )
    {        
        if ( widget != mainWidget )
            widget->GetParent()->BringToTop( widget );
        bool returnResult = widget->OnMouseDown( event.Mouse.Button,
                                                 event.Mouse.Modifiers, 
                                                 event.Mouse.x, 
                                                 event.Mouse.y );        
        return returnResult;
    }

        
    return false;
}

bool PawsManager::HandleMouseUp( iEvent &event )
{
    // Check to see if we are moving a widget.
    if ( movingWidget )
    {
        movingWidget = NULL;
        return true;
    }

    if ( resizingWidget )
    {
        resizingWidget->StopResize();
        resizingWidget = NULL;
        return true;
    }

    if ( currentFocusedWidget )
        return currentFocusedWidget->OnMouseUp( event.Mouse.Button, 
                                                event.Mouse.Modifiers,
                                                event.Mouse.x, 
                                                event.Mouse.y );

    
    pawsWidget* widget = mainWidget->WidgetAt( event.Mouse.x, event.Mouse.y );
    
    SetCurrentFocusedWidget ( widget );

    if ( widget )
    {
        if ( widget != mainWidget )
            widget->GetParent()->BringToTop( widget );

        return widget->OnMouseUp( event.Mouse.Button, 
                                  event.Mouse.Modifiers,
                                  event.Mouse.x, 
                                  event.Mouse.y );
    }

    return false;
}


bool PawsManager::HandleMouseMove( iEvent &event )
{  
    mouse->SetPosition( event.Mouse.x, event.Mouse.y );

    timeOver = csGetTicks();

    if ( movingWidget )
    {
        movingWidget->MoveDelta(mouse->GetDeltas().x, mouse->GetDeltas().y);
        return true;
    }
    
    if ( resizingWidget )
    {
        resizingWidget->Resize( mouse->GetDeltas().x, 
                                mouse->GetDeltas().y,
                                resizingFlags );
        return true;
    }

    if (modalWidget == NULL)
    {
        pawsWidget* widget = mainWidget->WidgetAt( event.Mouse.x, event.Mouse.y );

        if ( widget )
        {
            if ( widget && widget != mouseoverWidget )
            {
                widget->OnMouseEnter();
                if ( mouseoverWidget ) mouseoverWidget->OnMouseExit();
            }
            
            mouseoverWidget = widget;
            pawsWidget* widgetFade = widget;
            
            // Only fade in/out topmost parent
            if (widgetFade != mainWidget)
            {
                for (pawsWidget* mainParent = widgetFade->GetParent();mainParent != mainWidget;mainParent=widgetFade->GetParent())
                {
                    widgetFade = mainParent;
                }
            }                    
            
            
            if (lastfadeWidget != widgetFade && widgetFade && widgetFade->IsVisible())
            {               
                if (lastfadeWidget && lastfadeWidget != mainWidget)
                {
                    lastfadeWidget->MouseOver(false);                                                    
                }
                lastfadeWidget = widgetFade;
                
                if (widgetFade != mainWidget)
                    widgetFade->MouseOver(true);                // Fade in
            }            
        }
    }
    
    PositionDragDropWidget();

    return false;
}



void PawsManager::Draw()
{       
    mainWidget->DrawChildren();
    if ( modalWidget != NULL ) modalWidget->Draw();

    graphics2D->SetClipRect( 0,0, graphics2D->GetWidth(), graphics2D->GetHeight());

	// Draw the tooltip above all other windows after 1000 csTicks
	// Should be set in options
	if (timeOver < csGetTicks() - 1000 && mouseoverWidget)
		mouseoverWidget->DrawToolTip(mouse->GetPosition().x, mouse->GetPosition().y);

    if (dragDropWidget == NULL)
        mouse->Draw();
    else
        dragDropWidget->Draw();
}

void PawsManager::RegisterWidgetFactory( pawsWidgetFactory* fact )
{
    factories.Push( fact );
}

bool PawsManager::LoadWidget( const char* widgetFile )
{
    csString fullPath = localization->FindLocalizedFile(widgetFile);
    
    csRef<iDataBuffer> buff = vfs->ReadFile( fullPath );

    if ( !buff || !buff->GetSize() )
    {
        printf( "Could not load XML: %s", fullPath.GetData() );
        return false;
    }
    
    csRef<iDocument> doc = xml->CreateDocument();
    const char* error = doc->Parse( buff );
    if ( error )
    {
        printf( "Error parsing XML file %s: %s", fullPath.GetData(), error );
        return false;
    }

    
    csRef<iDocumentNode> root = doc->GetRoot();
    csRef<iDocumentNode> topNode = root->GetNode("widget_description");
    csRef<iDocumentNodeIterator> iter = topNode->GetNodes();
    
    while ( iter->HasNext() )
    {
        csRef<iDocumentNode> node = iter->Next();

        if ( node->GetType() != CS_NODE_ELEMENT )
            continue;


        // This is a widget so read it's factory to create it.
        if ( strcmp( node->GetValue(), "widget" ) == 0 )
        {
            csString factory = node->GetAttributeValue( "factory" );
            if ( factory.Length() == 0 )
            {
                printf("Could not read factory from XML file. Error in XML");
                return false;
            }
         
            pawsWidget* widget = CreateWidget( factory );
            if ( !widget )
            {
                printf( "Failed to create widget %s", factory.GetData() ); 
                return false;
            }

            // Add in to the main widget.
            mainWidget->AddChild( widget );            
            if ( widget->Load( node ) == false )
            {
                printf("Widget %s failed to load %s", widget->GetName(), factory.GetData() );
                return false;
            }
        }
    }
    
    return true;
}


pawsWidget* PawsManager::CreateWidget( const char* factoryName )
{
    csString factory( factoryName );    
    for ( int x = 0; x < factories.Length(); x++ )
    {                
        if ( factory == factories[x]->GetName() )
        {
            pawsWidget * newWidget = factories[x]->Create();
            newWidget->SetFactory( factoryName );
            return newWidget;
        }
    }

    printf("Could not locate Factory: %s", factoryName );
    return NULL;
}


void PawsManager::SetModalWidget( pawsWidget* widget )
{
    if (widget != NULL)
        SetCurrentFocusedWidget( widget );

    modalWidget = widget;
}


void PawsManager::SetCurrentFocusedWidget ( pawsWidget* widget )
{    
    if ( widget == NULL )
    {
        SetCurrentFocusedWidget( mainWidget );
        return;
    }

    if ( modalWidget == NULL )
    {
        if ( currentFocusedWidget != widget )
        {
            // Check if widget CAN be focused, if it can't, pass it to the parent
            for(;!(widget->OnGainFocus()) && widget != mainWidget;widget = widget->GetParent())
                if (widget == currentFocusedWidget) // Abort
                    break;
            if (currentFocusedWidget) currentFocusedWidget->OnLostFocus();

        }   
    
        currentFocusedWidget = widget; 
    }
    else
    {
        // Allow widget to be focused if it is a child of the modal widget
        pawsWidget* check = modalWidget->FindWidget( widget->GetName() );
        if (check)
            if ( currentFocusedWidget != widget )
            {
                // Check if widget CAN be focused, if it can't, pass it to the parent
                for(;!(widget->OnGainFocus()) && widget != mainWidget;widget = widget->GetParent())
                    if (widget == currentFocusedWidget) // Abort
                        break;
                widget->OnGainFocus();
                if (currentFocusedWidget) currentFocusedWidget->OnLostFocus();
            }
            currentFocusedWidget = widget; 
    }
}


void PawsManager::OnWidgetDeleted(pawsWidget * widget)
{
    if (currentFocusedWidget == widget)
        currentFocusedWidget = NULL;
    if (mouseoverWidget == widget)
        mouseoverWidget = NULL;
    if (movingWidget == widget)
        movingWidget = NULL;
    if (modalWidget == widget)
        modalWidget = NULL;
    if (resizingWidget == widget)
        resizingWidget = NULL;
    if ( lastfadeWidget == widget )
        lastfadeWidget = NULL;
}

csString PawsManager::Translate(const csString & orig)
{
    return localization->Translate(orig);
}
void PawsManager::CreateWarningBox( const char* message )
{
    pawsOkBox* Ok = (pawsOkBox*)FindWidget("OkWindow");
    if ( Ok )
    {
        Ok->SetText( message );
        Ok->MoveTo( (graphics2D->GetWidth() -  Ok->GetActualWidth(512) )/2 , 
                    (graphics2D->GetHeight() - Ok->GetActualHeight(256))/2);        
        Ok->Show();        
             
        SetModalWidget( Ok );                                        
    }   
}

pawsWidget * PawsManager::GetDragDropWidget()
{
    return dragDropWidget;
}

void PawsManager::SetDragDropWidget(pawsWidget * dragDropWidget)
{
    if (this->dragDropWidget != NULL)
        delete this->dragDropWidget;
    this->dragDropWidget = dragDropWidget;
    PositionDragDropWidget();
}

void PawsManager::PositionDragDropWidget()
{
    if (dragDropWidget != NULL)
    {
        csRect frame = dragDropWidget->ScreenFrame();
        csPoint mousePos = mouse->GetPosition();
        dragDropWidget->MoveTo(mousePos.x - frame.Width()/2, mousePos.y - frame.Height()/2);
    }
}

                
#define RegisterFactory(factoryclass)   \
    factory = new factoryclass(this);


void PawsManager::RegisterFactories()
{
    pawsWidgetFactory* factory;

    RegisterFactory (pawsBaseWidgetFactory);
//    RegisterFactory (pawsItemSlotFactory);
    RegisterFactory (pawsButtonFactory);
    RegisterFactory (pawsScrollBarFactory);    
    RegisterFactory (pawsObjectViewFactory);
    RegisterFactory (pawsTextBoxFactory);
    RegisterFactory (pawsMultiLineTextBoxFactory);
    RegisterFactory (pawsMessageTextBoxFactory);
    RegisterFactory (pawsEditTextBoxFactory);
//    RegisterFactory (pawsInfoWindowFactory);
    RegisterFactory (pawsYesNoBoxFactory);
    RegisterFactory (pawsListBoxFactory);
    RegisterFactory (pawsComboBoxFactory);
    RegisterFactory (pawsRadioButtonGroupFactory);
    RegisterFactory (pawsRadioButtonFactory);
    RegisterFactory (pawsProgressBarFactory);
    RegisterFactory (pawsOkBoxFactory);
    RegisterFactory (pawsSelectorBoxFactory);
/*    
    RegisterFactory (pawsSplashWindowFactory);
    RegisterFactory (pawsLoadWindowFactory);
    RegisterFactory (pawsChatWindowFactory);
    RegisterFactory (pawsInventoryWindowFactory);
    RegisterFactory (pawsItemDescriptionWindowFactory);
    RegisterFactory (pawsContainerDescWindowFactory);
    RegisterFactory (pawsInteractWindowFactory);
    RegisterFactory (pawsControlWindowFactory);
    RegisterFactory (pawsGroupWindowFactory);
    RegisterFactory (pawsExchangeWindowFactory);
    RegisterFactory (pawsSpellBookWindowFactory);
    RegisterFactory (pawsGlyphWindowFactory);
    RegisterFactory (pawsMerchantWindowFactory);
*/

    RegisterFactory (pawsSimpleTreeFactory);
    RegisterFactory (pawsSimpleTreeNodeFactory);
    RegisterFactory (pawsSeqTreeNodeFactory);
    RegisterFactory (pawsWidgetTreeNodeFactory);
/*
    RegisterFactory (pawsConfigWindowFactory);
    RegisterFactory (pawsConfigKeysFactory);
    RegisterFactory (pawsFingeringWindowFactory);
    RegisterFactory (pawsConfigDetailsFactory);
    RegisterFactory (pawsConfigMouseFactory);
    RegisterFactory (pawsConfigEntityLabelsFactory);

    RegisterFactory (pawsPetitionWindowFactory);
    RegisterFactory (pawsPetitionGMWindowFactory);
    RegisterFactory (pawsPetitionViewWindowFactory);
    RegisterFactory (pawsShortcutWindowFactory);
    RegisterFactory (pawsLoginWindowFactory);
    RegisterFactory (pawsCharacterPickerWindowFactory);
    RegisterFactory (pawsGuildWindowFactory);
    RegisterFactory (pawsGuildJoinWindowFactory);
    RegisterFactory (pawsLootWindowFactory);

    RegisterFactory (pawsCreationMainFactory);
    RegisterFactory (pawsCharParentsFactory);
    RegisterFactory (pawsChildhoodWindowFactory);    
    RegisterFactory (pawsLifeEventWindowFactory);    
    RegisterFactory (pawsPathWindowFactory);    
    RegisterFactory (pawsSummaryWindowFactory);    
*/    
    RegisterFactory (pawsMenuFactory);
    RegisterFactory (pawsMenuItemFactory);
    RegisterFactory (pawsMenuSeparatorFactory);
/*  RegisterFactory (pawsChatMenuItemFactory);

    RegisterFactory (pawsSkillWindowFactory);
    RegisterFactory (pawsQuestListWindowFactory);
    RegisterFactory (pawsSpellCancelWindowFactory);
*/  
    RegisterFactory (pawsCheckBoxFactory);
	RegisterFactory (pawsTabWindowFactory);
	//RegisterFactory (pawsQuitInfoBoxFactory);

}
/*
void PawsManager::CreateQuitInfoBox(const char *message)
{
	pawsQuitInfoBox* quitinfo = (pawsQuitInfoBox*)FindWidget("QuitInfoWindow");
	quitinfo->SetBox(message);
	quitinfo->Show();
}
*/
