/*
 * pawsmanager.h - Author: Andrew Craig
 *
 * Copyright (C) 2003 PlaneShift Team (info@planeshift.it,
 * http://www.planeshift.it)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation (version 2 of the License)
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */
// pawsmanager.h: interface for the PawsManager class.
//
//////////////////////////////////////////////////////////////////////
#ifndef PAWS_MANAGER_HEADER
#define PAWS_MANAGER_HEADER

#include <csutil/ref.h>
#include <csutil/parray.h>

#include <iutil/vfs.h>
#include <iutil/document.h>
//#include <physicallayer/entity.h>

#include "pawsmouse.h"
#include "psmousebinds.h"
//#include "net/msghandler.h"

struct iObjectRegistry;
struct iGraphics2D;
struct iGraphics3D;
struct iEvent;
class psLocalization;

class pawsWidget;
class pawsWidgetFactory;

class pawsMainWidget;
class pawsTextureManager;
class pawsPrefManager;

/** Main PlaneShift Window manager.
 */
class PawsManager
{
public:
    PawsManager(iObjectRegistry* objectReg);

    virtual ~PawsManager();

    void SetMainWidget(pawsMainWidget *widg);

    bool HandleEvent( iEvent& event );

    /** Draw the main widget and the mouse last.
     */
    void Draw();

    iGraphics2D* GetGraphics2D() { return graphics2D; }
    iGraphics3D* GetGraphics3D() { return graphics3D; }
    iObjectRegistry* GetObjectRegistry() { return objectReg; }

    pawsTextureManager* GetTextureManager() { return textureManager; }

    /** Add a new factory to the list that the manager knows about.
     * Each widget type must have it's own factory so the manager can
     * build it based on it's name.
     * @param factory The widget factory to add to the list.
     */
    void RegisterWidgetFactory( pawsWidgetFactory* factory );


    /** Loads a widget definition file.
     * This loads a widget from XML and adds it to the main widget.
     * @param widgetFile The standard path of the widget to load - see psLocalization::FindLocalizedFile()
     * @return True if the widget was loaded properly.
     */
    bool LoadWidget( const char* widgetFile );

    /** Create a new widget.
     * This creates a new widget based on the factory that is passed in.
     * @param factoryName The name of the factory that is used to create a
     *                    widget.
     * @return A new instance of a widget if the factory was found. NULL if the
               widget could not be found.
     */
    pawsWidget* CreateWidget( const char* factoryName );

    void SetCurrentFocusedWidget ( pawsWidget* widget );
    void SetModalWidget( pawsWidget* widget );

    /** pawsWidget destructor calls this so PawsManager can NULLify all its links to the widget
     */
    void OnWidgetDeleted(pawsWidget * widget);


    /** Let the window manager know that a widget is being moved.
     * @param moving The widget that is currently moving.
     */
    void MovingWidget( pawsWidget* moving );

    /** Let the manager know that a widget is being resized.
     * @param resize The widget that is being resized.
     * @param flags The resize flags that this widget should be resized with.
     */
    void ResizingWidget( pawsWidget* widget, int flags );

    /** Get the prefrence manager.
     */
    pawsPrefManager* GetPrefs() { return prefs; }

    pawsMouse* GetMouse() { return mouse; }

    pawsWidget* FindWidget( const char* name );

    pawsMainWidget * GetMainWidget() { return mainWidget; }

    // Retrieves psLocalization object:
    psLocalization * GetLocalization() { return localization; }

    // A shortcut - translation without need to call GetLocalization()
    csString Translate(const csString & orig);

    /** Gets and sets the widget that is being drag'n'dropped over screen with mouse
      * Parameter can be NULL. Ownership of the widget goes to psPawsManager.
      * This means that this widget must not be already owned (e.g. be child of another widget)
     */
    pawsWidget * GetDragDropWidget();
    void SetDragDropWidget(pawsWidget * dragDropWidget);
    
    void CreateWarningBox( const char* message );

	//void CreateQuitInfoBox( const char* message );

protected:
    /**
     * Sets position of dragDropWidget (if the is one) to position of mouse cursor
     */
    void PositionDragDropWidget();
    
    psLocalization * localization;

    /// The preference/default manager.
    pawsPrefManager * prefs;

    /// The last widget that the mouse clicked on. ( Hence the focused one ).
    pawsWidget* currentFocusedWidget;

    /// The last widget that the mouse moved over. ( For alpha fade effects ).
    pawsWidget* mouseoverWidget;

	/// The actual widget that was last faded
	pawsWidget* lastfadeWidget;

	/// The time mouse has been over the last widget
	csTicks timeOver;

    /// Current widget that is being moved.
    pawsWidget* movingWidget;

    /// Current Modal Widget
    pawsWidget* modalWidget;

    /// holds the current widget that is undergoing a resize.
    pawsWidget* resizingWidget;

    /// The resize flags.
    int resizingFlags;

    /// Helper function to load standard factories.
    void RegisterFactories();

    bool HandleMouseMove( iEvent& event );
    bool HandleMouseDown( iEvent& event );
    bool HandleMouseUp( iEvent& event );
    bool HandleKeyDown( iEvent& event );

    csRef<iObjectRegistry> objectReg;

    csRef<iGraphics2D> graphics2D;
    csRef<iGraphics3D> graphics3D;

    // The main handler widget.
    pawsMainWidget* mainWidget;

    // The texture manager
    pawsTextureManager* textureManager;

    // Mouse pointer
    pawsMouse* mouse;

    csPDelArray<pawsWidgetFactory> factories;

    csRef<iVFS> vfs;
    csRef<iDocumentSystem>  xml;
    
    bool hadKeyDown;

    /**
     * This is the widget that is drag'n'dropped across screen with mousse.
     * If it is not NULL, it is drawn instead of mouse on its position.
     * This widget has no parent widget (same as with pawsMainWidget)
     */    
    pawsWidget * dragDropWidget;
};


#endif


