/*
 * pawsobjectview.cpp - Author: Andrew Craig
 *
 * Copyright (C) 2003 PlaneShift Team (info@planeshift.it,
 * http://www.planeshift.it)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation (version 2 of the License)
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */

#include <cssysdef.h>
#include <iutil/object.h>
#include <iutil/objreg.h>
#include <imap/parser.h>
#include <iutil/vfs.h>

#include "pawsobjectview.h"
#include "pawsmanager.h"
#include "util/log.h"
#include "util/psconst.h"

int pawsObjectView::idName = 0;

pawsObjectView::pawsObjectView( PawsManager* manager )
               :pawsWidget( manager )
{
    engine = CS_QUERY_REGISTRY( windowManager->GetObjectRegistry(), iEngine );
    object = NULL;
    idName++;  
    
    loadedMap = false;      
}   

pawsObjectView::~pawsObjectView()
{
    idName--;
}

bool pawsObjectView::Setup(iDocumentNode* node )
{
    csRef<iDocumentNode> mapNode = node->GetNode( "map" ); 
    if ( mapNode )
    {
        csString mapFile = mapNode->GetAttributeValue("file");
        csString sector  = mapNode->GetAttributeValue("sector");
       
        return LoadMap( mapFile, sector );
    }
    else
    {
        return CreateMap();
    }
}

bool pawsObjectView::LoadMap( const char* map, const char* sector )
{
    csRef<iEngine> engine = CS_QUERY_REGISTRY( windowManager->GetObjectRegistry(), iEngine );
    csRef<iLoader> loader = CS_QUERY_REGISTRY( windowManager->GetObjectRegistry(), iLoader );
    csRef<iVFS> VFS = CS_QUERY_REGISTRY( windowManager->GetObjectRegistry(), iVFS);
             
    char newName[10];
    sprintf(newName, "NAME%d\n", idName );
    
    stage = engine->FindSector( sector );
    if ( !stage )
    {   
        iRegion* cur_region = engine->CreateRegion (newName);

        // Clear it out if it already existed
        cur_region->DeleteAll ();
    
        // Now load the map into the selected region
        VFS->ChDir (map);
        engine->SetCacheManager(NULL);
        if ( !loader->LoadMapFile("world", CS_LOADER_KEEP_WORLD, cur_region, CS_LOADER_ACROSS_REGIONS) )
            return false;
    
        stage = engine->FindSector( sector );
        CS_ASSERT( stage );
        cur_region->Add( stage->QueryObject() );
        if ( !stage )
             return false;
            
        cur_region->Prepare();            
    }
    
    meshSector = engine->CreateSector( newName ); 
        
    iLightList* lightList = meshSector->GetLights();
    csRef<iLight> light = engine->CreateLight(NULL, csVector3(-3,4,-3),10,
                                      csColor(0.86F,0.87F,0.6F), CS_LIGHT_STATIC);    
    light->SetAttenuation( CS_ATTN_NONE );
    lightList->Add( light );
       
    meshSector->ShineLights();
    
    meshView = csPtr<iView> (new csView( engine, windowManager->GetGraphics3D() ));
    meshView->GetCamera()->SetSector(meshSector);
    meshView->GetCamera()->GetTransform().SetOrigin(csVector3(0,1,-4));

    meshView->SetRectangle(screenFrame.xmin, screenFrame.ymin,
                       screenFrame.Width(),screenFrame.Height());
           
    view = csPtr<iView> (new csView( engine, windowManager->GetGraphics3D() ));
    view->GetCamera()->SetSector(stage);
    view->GetCamera()->GetTransform().SetOrigin(csVector3(0,1,-4));
    
    
    view->SetRectangle(screenFrame.xmin, screenFrame.ymin,
                       screenFrame.Width(),screenFrame.Height());
   
    loadedMap = true;        
    return true;        
}

bool pawsObjectView::CreateMap()
{
    csRef<iEngine> engine = CS_QUERY_REGISTRY( windowManager->GetObjectRegistry(), iEngine );
    csRef<iLoader> loader = CS_QUERY_REGISTRY( windowManager->GetObjectRegistry(), iLoader );
    csRef<iVFS> VFS = CS_QUERY_REGISTRY( windowManager->GetObjectRegistry(), iVFS);
    
    char newName[10];
    sprintf(newName, "NAME%d\n", idName );
    iRegion* cur_region = engine->CreateRegion (newName);

    // Clear it out if it already existed
    cur_region->DeleteAll ();
       iTextureWrapper* txt = loader->LoadTexture("stone", "/lib/std/stone4.gif");
       
    if ( !txt )
    {
         Error1("Error loading the object view texture");
        return false;
    }
    cur_region->Add( txt->QueryObject () );
    
    iMaterialWrapper * matWrap;
    matWrap = engine->GetMaterialList()->FindByName( "stone" );

    cur_region->Add( matWrap->QueryObject() );

    stage = engine->CreateSector( "stage" );
    cur_region->Add( stage->QueryObject() );

    
    csRef<iMeshWrapper> walls = engine->CreateSectorWallsMesh( stage, "stagewalls" );
    cur_region->Add( walls->QueryObject() );


    csRef<iThingState> ws = SCF_QUERY_INTERFACE (walls->GetMeshObject (), iThingState);
    csRef<iThingFactoryState> walls_state = ws->GetFactory ();
    walls_state->AddInsideBox(csVector3 (-5, 0, -5), csVector3 (5, 10, 5));
    walls_state->SetPolygonMaterial(CS_POLYRANGE_LAST, matWrap);
    walls_state->SetPolygonTextureMapping (CS_POLYRANGE_LAST, 3);
    
    // Create some lights. Later this can be added using the .def file or
    // a world file. 
    iLightList* lightList = stage->GetLights();
    csRef<iLight> light = engine->CreateLight(NULL, csVector3(-3,4,-3),10,
                                      csColor(0.86F,0.87F,0.6F), CS_LIGHT_STATIC);    
    light->SetAttenuation( CS_ATTN_NONE );
    lightList->Add( light );
    cur_region->Add( light->QueryObject() );

    light = engine->CreateLight(NULL, csVector3 (3, 4, 2), 10,
                                csColor (0, .125F, .85F), CS_LIGHT_STATIC);    
    lightList->Add (light);
    cur_region->Add( light->QueryObject() );
    
    light = engine->CreateLight(NULL, csVector3(3,1,-3),10,
                                csColor(0.45F,0.45F,0.45F), CS_LIGHT_STATIC);    
    lightList->Add( light );
    cur_region->Add( light->QueryObject() );
    
    // Cannot Prepare() entire engine more than once
    cur_region->Add( stage->QueryObject() );
    cur_region->Prepare();
    engine->Prepare(); 
    

    view = csPtr<iView> (new csView( engine, windowManager->GetGraphics3D() ));
    view->GetCamera()->SetSector(stage);
    view->GetCamera()->GetTransform().SetOrigin(csVector3(0,1,-4));

    view->SetRectangle(screenFrame.xmin, screenFrame.ymin,
                       screenFrame.Width(),screenFrame.Height());
                       
    view->GetCamera()->GetTransform().SetOrigin(csVector3(0,1,-4));

    return true; 
}

void pawsObjectView::View( const char* factName, const char* fileName )
{        
    if ( object )
    {
        object->GetMovable()->ClearSectors();
        engine->GetMeshes()->Remove(object);
    }    
    
    csRef<iMeshFactoryWrapper> meshfact = 0;
    meshfact = engine->GetMeshFactories()->FindByName (factName);
            
    if ( !meshfact )
    {
        csRef<iLoader> loader = CS_QUERY_REGISTRY (windowManager->GetObjectRegistry(), iLoader);
        meshfact = loader->LoadMeshObjectFactory (fileName);                
    }

    iSector* sector =  loadedMap ? meshSector : stage;

    object = engine->CreateMeshWrapper (meshfact, "PaperDoll", sector, csVector3(0,0,0) );
    
//    object->DeferUpdateLighting( CS_NLIGHT_DYNAMIC | CS_NLIGHT_STATIC, 10 ); 
    
    csBox3 box;
    object->GetWorldBoundingBox( box );
    objectPos = box.GetCenter();
    
    
}

void pawsObjectView::View( iMeshWrapper* wrapper )
{
    iSector* sector =  loadedMap ? meshSector : stage;
    object = engine->CreateMeshWrapper ( wrapper->GetFactory(), "PaperDoll", sector, csVector3(0,0,0) );

//    object->DeferUpdateLighting( CS_NLIGHT_DYNAMIC | CS_NLIGHT_STATIC, 10 ); 
    
    csBox3 box;
    object->GetWorldBoundingBox( box );
    objectPos = box.GetCenter();
}

void pawsObjectView::Draw()
{
    graphics2D->SetClipRect( 0,0, graphics2D->GetWidth(), graphics2D->GetHeight());         
    if ( !windowManager->GetGraphics3D()->BeginDraw( engine->GetBeginDrawFlags() | CSDRAW_3DGRAPHICS) )
        return;

    if ( !view )
        return;

    iGraphics3D* og3d = view->GetContext();

    view->SetContext(windowManager->GetGraphics3D());

    view->SetRectangle( screenFrame.xmin, 
                       windowManager->GetGraphics3D()->GetHeight() - screenFrame.ymax ,
                       screenFrame.Width(),
                       screenFrame.Height() );
                                                        
    view->GetCamera()->SetPerspectiveCenter(
                       screenFrame.xmin + (screenFrame.Width() >> 1),
                       windowManager->GetGraphics3D()->GetHeight() - screenFrame.Height() - 
                       screenFrame.ymin + (screenFrame.Height() >> 1) );

    
    view->GetCamera()->SetFOV( view->GetCamera()->GetFOV(), screenFrame.Width() );

        
    csVector3 orig = view->GetCamera()->GetTransform().GetOrigin();

    view->GetCamera()->GetTransform().LookAt( objectPos - orig, csVector3(0,1,0) );       
    view->Draw();
    
    if ( loadedMap )
    {
        og3d = meshView->GetContext();

        meshView->SetContext(windowManager->GetGraphics3D());

        meshView->SetRectangle( screenFrame.xmin, 
                                windowManager->GetGraphics3D()->GetHeight() - screenFrame.ymax ,
                                screenFrame.Width(),
                                screenFrame.Height() );
                                                        
        meshView->GetCamera()->SetPerspectiveCenter(
                                screenFrame.xmin + (screenFrame.Width() >> 1),
                                windowManager->GetGraphics3D()->GetHeight() - screenFrame.Height() - 
                                screenFrame.ymin + (screenFrame.Height() >> 1) );

    
        meshView->GetCamera()->SetFOV( view->GetCamera()->GetFOV(), screenFrame.Width() );
    
        orig = meshView->GetCamera()->GetTransform().GetOrigin();
        meshView->GetCamera()->GetTransform().LookAt( objectPos - orig, csVector3(0,1,0) );    
        meshView->Draw();        
    }   
    
    windowManager->GetGraphics3D()->BeginDraw( CSDRAW_2DGRAPHICS );

    view->SetContext( og3d );
    pawsWidget::Draw();
}
