/*
 * pawsprefmanager.cpp - Author: Andrew Craig
 *
 * Copyright (C) 2003 PlaneShift Team (info@planeshift.it,
 * http://www.planeshift.it)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation (version 2 of the License)
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */
// pawsprefmanager.cpp: implementation of the pawsPrefManager class.
//
//////////////////////////////////////////////////////////////////////

#include <cssysdef.h>
#include "pawsprefmanager.h"
#include "pawsmanager.h"
#include "pawsborder.h"

#include <csutil/util.h>

#include <csutil/xmltiny.h>
#include <csutil/scfstr.h>
#include <iutil/objreg.h>
#include <iutil/vfs.h>
#include <iutil/document.h>
#include <ivideo/fontserv.h>
#include <ivideo/graph2d.h>
#include "util/log.h"


//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

pawsPrefManager::pawsPrefManager( PawsManager* mgr )
{
    manager = mgr;
    objectReg = manager->GetObjectRegistry();

    vfs = CS_QUERY_REGISTRY( objectReg, iVFS );
    xml = csPtr<iDocumentSystem> ( new csTinyDocumentSystem ); 
    graphics2D = CS_QUERY_REGISTRY( objectReg, iGraphics2D );
}


pawsPrefManager::~pawsPrefManager()
{

}

bool pawsPrefManager::LoadPrefFile( const char* fileName )
{
    csRef<iDataBuffer> buf = vfs->ReadFile( fileName );
    
    if ( !buf || !buf->GetSize() )
    {
        return false;
    }

    csRef<iDocument> doc = xml->CreateDocument();
    const char* error = doc->Parse( buf );
    if ( error )
        printf("ERROR: %s\n", error);

    csRef<iDocumentNode> root = doc->GetRoot();

    csRef<iDocumentNode> topNode = root->GetNode("prefs");

    csRef<iDocumentNodeIterator> iter = topNode->GetNodes();

    while ( iter->HasNext() )
    {
        csRef<iDocumentNode> node = iter->Next();
        
        // default font here
        if ( strcmp( node->GetValue(), "font" ) == 0 )
        {
            csString name = node->GetAttributeValue( "name" );
            int size = node->GetAttributeValueAsInt("size");
            defaultFont = graphics2D->GetFontServer()->LoadFont( name, (size)?size:10 );
            if ( !defaultFont )
            {
                Error2("Could not load font: >%s<", (const char*)name );
                return false;
            }

            int r = node->GetAttributeValueAsInt( "r" );            
            int g = node->GetAttributeValueAsInt( "g" );            
            int b = node->GetAttributeValueAsInt( "b" );            

            defaultFontColour = graphics2D->FindRGB( r, g, b );
        }

        if ( strcmp( node->GetValue(), "border" ) == 0 )
        {
            LoadBorderColours( node );
        }
 
        if ( strcmp( node->GetValue(), "chatcolours" ) == 0 )
        {
            LoadChatColours( node );
        }
         
    }
    
    return true;
}


bool pawsPrefManager::LoadBorderFile( const char* file )
{   
    // A list of the XML tags for borders.
    static const char* borderTags[] = 
    {
     "topleft",
     "topright",
     "bottomleft", 
     "bottomright",
     "leftmiddle",
     "rightmiddle",
     "topmiddle",
     "bottommiddle",
     0
     };
    
      
    csRef<iDataBuffer> buf = vfs->ReadFile( file );
    
    if ( !buf || !buf->GetSize() )
    {
        return false;
    }
    
    csRef<iDocument> doc = xml->CreateDocument();
    const char* error = doc->Parse( buf );
    if ( error )
        printf("ERROR: %s\n", error);
    csRef<iDocumentNode> root = doc->GetRoot();
    csRef<iDocumentNode> topNode = root->GetNode("borderlist");
    csRef<iDocumentNodeIterator> iter = topNode->GetNodes();
    
    while ( iter->HasNext() )
    {
        csRef<iDocumentNode> node = iter->Next();        
        
        BorderDefinition* definition = new BorderDefinition;
        definition->name = node->GetAttributeValue("name");
        int current = 0;
        // For each border tag ( in the static above ) get that node tag
        // and construct an image description from it. 
        while ( borderTags[current] != 0 )
        {
            csRef<iDocumentNode> tempnode = node->GetNode( borderTags[current] );        
            csRef<iDocumentNode> image = tempnode->GetNode("image");                        
            manager->GetTextureManager()->ReadImageDescription( image );
            
            definition->descriptions[current] = image->GetAttributeValue("resource");
                                                                             
            current++;
        }
        
        borders.Push( definition );                        
    }
    return true;
}


BorderDefinition* pawsPrefManager::GetBorderDefinition( const char* name )
{
    for ( int x = 0; x < borders.Length(); x++ )
    {
        if ( borders[x]->name == name )
            return borders[x];
    }
       
    return 0;
}


void pawsPrefManager::LoadChatColours( iDocumentNode* node )
{
    csRef<iDocumentNodeIterator> iter = node->GetNodes();
    
    while ( iter->HasNext() )
    {
        csRef<iDocumentNode> colour = iter->Next();
        
        int r = colour->GetAttributeValueAsInt( "r" );            
        int g = colour->GetAttributeValueAsInt( "g" );            
        int b = colour->GetAttributeValueAsInt( "b" );            
        
        int col = graphics2D->FindRGB( r, g, b );

        csString nodeName ( colour->GetValue() );
        if ( nodeName == "systemtext" ) systemColour = col;
        if ( nodeName == "admintext"  ) adminColour = col;
        if ( nodeName == "playernametext" ) nameSaidColour = col;
        if ( nodeName == "chattext") chatColour = col;
        if ( nodeName == "telltext") tellColour = col;
        if ( nodeName == "shouttext") shoutColour = col;
        if ( nodeName == "guildtext") guildColour = col;
        if ( nodeName == "yourtext") yourColour = col;
        if ( nodeName == "grouptext") groupColour = col;
        if ( nodeName == "auctiontext") auctionColour = col;

    }
}


void pawsPrefManager::LoadBorderColours( iDocumentNode* node )
{
    csRef<iDocumentNodeIterator> iter = node->GetNodes();
    
    int index = 0;
    while ( iter->HasNext() )
    {
        csRef<iDocumentNode> colour = iter->Next();
        
        int r = colour->GetAttributeValueAsInt( "r" );            
        int g = colour->GetAttributeValueAsInt( "g" );            
        int b = colour->GetAttributeValueAsInt( "b" );            
        
        int col = graphics2D->FindRGB( r, g, b );
        borderColours[index++] = col;
    }    
}

