/*
 * pawswidget.cpp - Author: Andrew Craig
 *
 * Copyright (C) 2003 PlaneShift Team (info@planeshift.it,
 * http://www.planeshift.it)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation (version 2 of the License)
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */
// pawswidget.cpp: implementation of the pawsWidget class.
//
//////////////////////////////////////////////////////////////////////
#include <cssysdef.h>

#include <iutil/cfgmgr.h>
#include <iutil/evdefs.h>
#include <ivideo/fontserv.h>
#include <csutil/xmltiny.h>

#include "localization.h"

#include "pawswidget.h"
#include "pawsmanager.h"
#include "pawstexturemanager.h"
#include "pawsprefmanager.h"
#include "pawsborder.h"
#include "pawsmenu.h"
#include "pawsmainwidget.h"
#include "util/log.h"
#include "util/psstring.h"
#include "widgetconfigwindow.h"

#define RX(x) ((int) ((((float)x)*ratioX)+.5))
#define RY(y) ((int) ((((float)y)*ratioY)+.5))

#define PAWS_CONSTRUCTION 1

#define DEFAULT_FONT_SIZE   12
#define DFFAULT_FONT_STYLE   0
#define FONT_DROPSHADOW      1

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

pawsWidget::pawsWidget( PawsManager* manager )
{
    windowManager = manager;
    graphics2D = windowManager->GetGraphics2D();

    screenFrame = csRect(0,0,0,0);
    clipRect = csRect(0,0,0,0);
    saveWidgetPositions = false;

    visible  = true;
    movable  = false;
    alwaysOnTop = false;
    isResizable = false;
    id = -1;
    parent   = NULL;
    bgColour = -1;
    bgImage  = NULL;
    border   = NULL;
    name = "None";
    attachFlags = 0;
    fadeVal = 100;
    fadeSpeed = 1.0f;
    fade = true;
    alphaMin=0;
    //alphaMax=0;
    hasFocus = false;
    hasMouseFocus=false;
    configurable = false;
    fontStyle = 0;
    min_height = DEFAULT_MIN_HEIGHT;
    min_width = DEFAULT_MIN_WIDTH;
    max_height = graphics2D->GetHeight();
    max_width = graphics2D->GetWidth();
    hasBorderColours = false;
    contextMenu = NULL;
    xmlbinding = "";
}

pawsWidget::~pawsWidget()
{
    // Check to see if the widget position should be save.
    if ( saveWidgetPositions )
        SavePosition();

    if ( configurable )
        SaveSettings();

    while ( children.Length() > 0 )
    {
        delete children.Pop();
    }
    
    if (contextMenu != NULL)
    {
        windowManager->GetMainWidget()->DeleteChild(contextMenu);
        contextMenu = NULL;
    }
    
    windowManager->OnWidgetDeleted(this);

    delete bgImage;
    delete border;
}

bool pawsWidget::CheckButtonPressed( int button, int modifiers, pawsWidget* pressedWidget )
{  
    if ( pressedWidget->GetName()!=NULL  &&  GetCloseName()!=NULL  &&  strcmp( pressedWidget->GetName(),GetCloseName())==0 )
    {
        Close();
        return true;
    }
    else
        return OnButtonPressed( button, modifiers, pressedWidget );
}                 

/*
 * Check to see if this widget contains these coordinates at all.
 * Recurse through children to see if they could the one the mouse is over.
 * Return this widget if no children contain the coords.
 */
pawsWidget* pawsWidget::WidgetAt( int x, int y )
{      
    if ( Contains( x, y ) == false )
    {
        return NULL;
    }

    //Recurse through children to see if they could the one the mouse is over.
    for( int z = 0; z < children.Length(); z++ )
    {
        if ( children[z]->IsVisible() )
        { 
            if ( children[z]->Contains(x,y) )
                return children[z]->WidgetAt( x , y );
        }
    }
        
    //Return this widget if no children contain the coords.
    return this;
}

bool pawsWidget::Contains( int x, int y )
{
    if ( border )
    {
        return border->GetRect().Contains(x,y);
    }           
    else
        return screenFrame.Contains(x,y);       
}

void pawsWidget::SetParent( pawsWidget* newParent )
{
    parent = newParent;
}

void pawsWidget::AddChild( pawsWidget* childWidget )
{
    if ( !childWidget )
        return;

    if ( childWidget->IsAlwaysOnTop() )
        children.Insert( 0, childWidget );
    else
        children.Push( childWidget );

    // Let the child know that he is attached to this widget.
    childWidget->SetParent( this );
}

void pawsWidget::RemoveChild( pawsWidget* widget )
{
    if ( !widget )
        return;

    children.Delete( widget );   
}

void pawsWidget::DeleteChild( pawsWidget* widget )
{
    RemoveChild(widget);
    delete widget;
}

bool pawsWidget::LoadAttributes( iDocumentNode* node )
{
    // Get the name for this widget.
    csRef<iDocumentAttribute> atr = node->GetAttribute("name");
    if ( !atr )
    {
        printf("Widget has not defined a name. Error in XML");
        return false;
    }
    SetName(csString( atr->GetValue() ));


    // Check to see if this widget is visible directly after a load.
    atr = node->GetAttribute( "visible" );
    if ( atr )
    {
        csString choice = csString( atr->GetValue() );
        if ( choice == "yes" ) Show();
        if ( choice == "no"  ) Hide();
    }
    

    // Check to see if this widget should save it's position
    atr = node->GetAttribute("savepositions");
    if ( atr )
    {
        csString choice = csString(atr->GetValue());
        if ( choice == "yes" ) saveWidgetPositions = true;
        if ( choice == "no"  ) saveWidgetPositions = false;        
    }

    // Check to see if this widget is movable
    atr = node->GetAttribute("movable");
    if ( atr )
    {
        csString choice = csString(atr->GetValue());
        if ( choice == "yes" ) movable = true;
        if ( choice == "no"  ) movable = false;        
    }

    // Check to see if this widget is configurable
    atr = node->GetAttribute("configurable");
    if ( atr )
    {
        csString choice = csString(atr->GetValue());
        if ( choice == "yes" ) configurable = true;
        if ( choice == "no"  ) configurable = false;        
    }

    atr = node->GetAttribute( "resizable" );
    if ( atr )
    {
        csString choice = csString(atr->GetValue());
        if ( choice == "yes" ) isResizable = true;
        if ( choice == "no"  ) isResizable = false;        
    }
    
    atr = node->GetAttribute( "ContextMenu" );
    if ( atr )
        contextMenuFile = atr->GetValue();

    atr = node->GetAttribute( "id" );
    if ( atr )
    {
        id = atr->GetValueAsInt();
    }

    atr = node->GetAttribute( "xmlbinding" );
    if ( atr )
    {
        xmlbinding = atr->GetValue();
    }

	
	// Get tool tip, if any
    atr = node->GetAttribute( "tooltip" );
    if (atr)
    {
        toolTip = atr->GetValue();
    }

    // Get the optional font for this widget
    csRef<iDocumentNode> fontAttribute = node->GetNode("font");
    if ( fontAttribute )
    {
            csString name = fontAttribute->GetAttributeValue( "name" );
            int size = fontAttribute->GetAttributeValueAsInt("size");

            myFont = graphics2D->GetFontServer()->LoadFont( name , (size)? size:10 );
            if ( !myFont )
            {
                printf("Could not load font: >%s<", (const char*)name );
                return false;
            }
            int r = fontAttribute->GetAttributeValueAsInt( "r" );            
            int g = fontAttribute->GetAttributeValueAsInt( "g" );            
            int b = fontAttribute->GetAttributeValueAsInt( "b" );            

            defaultFontColour = graphics2D->FindRGB( r, g, b );

            r = fontAttribute->GetAttributeValueAsInt( "sr" );            
            g = fontAttribute->GetAttributeValueAsInt( "sg" );            
            b = fontAttribute->GetAttributeValueAsInt( "sb" );

            defaultFontShadowColour = graphics2D->FindRGB( r,g,b);
            
            if (fontAttribute->GetAttributeValueAsBool( "shadow" ))
                fontStyle |= FONT_DROPSHADOW;
    }


    // Get the frame for this widget.
    csRef<iDocumentNode> frameNode = node->GetNode( "frame" );    
    if ( frameNode )
    {
        defaultFrame.xmin = GetActualWidth(frameNode->GetAttributeValueAsInt("x"));
        defaultFrame.ymin = GetActualHeight(frameNode->GetAttributeValueAsInt("y"));
        int width  = GetActualWidth(frameNode->GetAttributeValueAsInt("width"));
        int height = GetActualHeight(frameNode->GetAttributeValueAsInt("height"));
        
        if(width < min_width)
            width = min_width;
        if(height < min_height)
            height = min_height;
        
        defaultFrame.SetSize( width, height );

        // If this widget has a parent then move it to it's correct relative position.
        if ( parent )
        {
            MoveTo(parent->ScreenFrame().xmin + defaultFrame.xmin,
                   parent->ScreenFrame().ymin + defaultFrame.ymin);
            screenFrame.SetSize( width, height );
        }
        else
        {
            MoveTo( defaultFrame.xmin, defaultFrame.ymin );
            SetSize( defaultFrame.Width(), defaultFrame.Height() );
        }
        
        csRef<iDocumentAttribute> useBorder = frameNode->GetAttribute("border");
        if ( useBorder )
        {
            csString borderString = useBorder->GetValue();
           
            if ( borderString != "no" )
            {
                if ( borderString == "yes" )
                    borderString.Replace("line");            
                
                border = new pawsBorder( WindowManager(), borderString );
                border->SetParent(this);                 
            }                          
        }
        
    }   

    // Process title bar, if any
    csRef<iDocumentNode> titleNode = node->GetNode("title");
    if (titleNode)
    {
        csString title = titleNode->GetAttributeValue("text");
        csString image = titleNode->GetAttributeValue("resource");               
        csString align = titleNode->GetAttributeValue("align");        
        csString close = titleNode->GetAttributeValue( "close_button" );
        
        SetTitle( title, image, align, close ); 
    }

    csRef<iDocumentNode> borderNode = node->GetNode( "childborders" );    
    if (borderNode)
        LoadBorderColours(borderNode);

    csRef<iDocumentNode> bgColourNode = node->GetNode( "bgcolour" );    
    if ( bgColourNode )
    {
        int r = bgColourNode->GetAttributeValueAsInt("r");
        int g = bgColourNode->GetAttributeValueAsInt("g");     
        int b = bgColourNode->GetAttributeValueAsInt("b");

        bgColour = graphics2D->FindRGB( r, g, b );
    }
    else
    {
        bgColour = -1;
    }
    
    // Get the background image attribute
    csRef<iDocumentNode> bgImageNode = node->GetNode( "bgimage" );    
    if ( bgImageNode )
    {
        csString image = bgImageNode->GetAttributeValue("resource");               
        psImageDescription* imageDesc = windowManager->GetTextureManager()->GetImageDescription( image );        
        bgImage = new pawsImage( windowManager->GetObjectRegistry() );
        bgImage->Description( imageDesc );

        csRef<iDocumentAttribute> alphaattr = bgImageNode->GetAttribute("alpha");
        if( alphaattr )    
            bgImage->SetAlpha( alphaattr->GetValueAsInt() );
        // Get the initial alpha value
        alpha = bgImage->GetAlpha();
		//alphaMax = alpha;
        csRef<iDocumentAttribute> fadeattr = bgImageNode->GetAttribute("fading");
		if (fadeattr)
			fade = fadeattr->GetValueAsBool();
    }
    else
    {
        bgImage = NULL;
    }

	if ( configurable )
	{
		LoadSettings();
	};

    // Get the attachment points
    csRef<iDocumentNode> attachPointsNode = node->GetNode( "attachpoints" );   
    if ( attachPointsNode )
    {
        csRef<iDocumentNodeIterator> points = attachPointsNode->GetNodes();
        while ( points->HasNext() )
        {
            csRef<iDocumentNode> point = points->Next();
            attachFlags |= GetAttachFlag( point->GetAttributeValue("point") );                                              
        }            
    }

    // Get the min width and height
    csRef<iDocumentNode> minFrameNode = node->GetNode( "minframe" );
    if ( minFrameNode )
    {
        min_width = GetActualWidth(minFrameNode->GetAttributeValueAsInt("width"));
        min_height = GetActualHeight(minFrameNode->GetAttributeValueAsInt("height"));
    }

    // Get the max width and height
    csRef<iDocumentNode> maxFrameNode = node->GetNode( "maxframe" );
    if ( maxFrameNode )
    {
        max_width = GetActualWidth(maxFrameNode->GetAttributeValueAsInt("width"));
        max_height = GetActualHeight(maxFrameNode->GetAttributeValueAsInt("height"));
    }
    
    // Check to see if we have possible saved positions to use. 
    if ( saveWidgetPositions )
    {
        csRect pos = LoadPositions();
        // Resize to loaded positions.
        MoveTo( pos.xmin, pos.ymin );
       
        //only set size if widget is resizable
        if( IsResizable() )
            SetSize( pos.Width(), pos.Height() );        
    }

    return true;
}

bool pawsWidget::Load( iDocumentNode* node )
{
    if ( ! LoadAttributes(node) )
        return false;

    // Call the setup for this widget.     
    if ( !Setup( node ) ) 
    {
        printf("Widget %s Setup Failed", GetName() );
        return false;
    }

    if ( ! LoadChildren(node) )
        return false;
        
    if ( !PostSetup() )
    {
        printf("Failed the PostSetup for %s", GetName() );
        return false;
    }
    
    Resize();
    StopResize();
    return true;
}

bool pawsWidget::LoadChildren( iDocumentNode* node )
{
    // Get an iterator over all the child widgets       
    csRef<iDocumentNodeIterator> childIter = node->GetNodes ("widget");

    while ( childIter->HasNext() )
    {
        csRef<iDocumentNode> childWidgetNode = childIter->Next();
        csString factory = childWidgetNode->GetAttributeValue( "factory" );

        if ( factory.Length() == 0 )
        {
            printf("Could not read factory from XML file. Error in XML");
            return false;
        }
                    
        pawsWidget* widget = windowManager->CreateWidget( factory );
        CS_ASSERT( widget );

        // Add in to the main widget.
        AddChild( widget );            
        if ( widget->Load( childWidgetNode ) == false )
        {
            printf("Failed to load child widget");
            return false;
        }
    }

    return true;
}

void pawsWidget::Show()
{
    visible = true;
    if ( border ) border->Show();    
    BringToTop( this );
}
    
void pawsWidget::Hide()
{ 
    visible = false; 
    if ( border ) 
        border->Hide();
}    

void pawsWidget::SetRelativeFrame( int x, int y, int width, int height )
{
    SetRelativeFramePos( x, y );
    SetRelativeFrameSize( width, height );
}

void MoveRect(csRect & rect, int x, int y)
{
    int width  = rect.Width();
    int height = rect.Height();
    
    rect.Set(x, y, x+width, y+height);
}

void pawsWidget::SetRelativeFramePos( int x, int y )
{
    MoveRect(defaultFrame, x, y);
    if (parent != NULL)
        MoveRect(screenFrame, parent->ScreenFrame().xmin + x, parent->ScreenFrame().ymin + y);
    else
        MoveRect(screenFrame, y, y);
    
    for ( int x = 0; x < children.Length(); x++ )        
        children[x]->RecalcScreenPositions();         
}

void pawsWidget::SetRelativeFrameSize( int width, int height )
{
    defaultFrame .SetSize( width, height );
    screenFrame  .SetSize( width, height );

    OnResize();
    
    for ( int x = 0; x < children.Length(); x++ )        
        children[x]->RecalcScreenPositions();         
}

void pawsWidget::RecalcScreenPositions()
{
    if ( parent )
    {
        screenFrame.xmin  = parent->ScreenFrame().xmin + defaultFrame.xmin;
        screenFrame.ymin  = parent->ScreenFrame().ymin + defaultFrame.ymin;
        screenFrame.SetSize( defaultFrame.Width(), defaultFrame.Height() );

        for ( int x = 0; x < children.Length(); x++ )        
            children[x]->RecalcScreenPositions();         
    }
}
    
void pawsWidget::UseBorder( const char* style )
{
    border = new pawsBorder( WindowManager(), style );
    border->SetParent( this );    
}

void pawsWidget::SetBackground( const char* image )
{
    psImageDescription* imageDesc = windowManager->GetTextureManager()->GetImageDescription( image );

    if ( imageDesc == NULL )
    {
        Warning2( LOG_PAWS, "Could Not load image: %s\n", image );
        return;
    }

   if ( bgImage )
        delete bgImage;

    bgImage = new pawsImage( windowManager->GetObjectRegistry() );

    bgImage->Description( imageDesc );
    // Get the initial alpha value
    alpha = bgImage->GetAlpha();
}

csString pawsWidget::GetBackground()
{
    if (bgImage && bgImage->GetDescription() )
    {
        return bgImage->GetDescription()->resourceName;
    }

    return csString("");
}


void pawsWidget::SetBackgroundAlpha( int value )
{
    if ( bgImage )
    {
        bgImage->SetAlpha( value );
        // Get the initial alpha value
        alpha = value;
		alphaMin = value;
    }
}


pawsWidget* pawsWidget::FindWidget( const char* name )
{
    csString temp(name);
    if ( this->name == temp )
        return this;

    for ( int z = 0; z < children.Length(); z++ )
    {
         pawsWidget* widget = children[z]->FindWidget( name );
         if ( widget != NULL )
             return widget;
    }

    //Warning2( LOG_PAWS, "Could not locate widget %s", name );           
    return NULL;
}


pawsWidget* pawsWidget::FindWidget( int ID )
{
    if ( this->id == ID )
        return this;

    for ( int z = 0; z < children.Length(); z++ )
    {
         pawsWidget* widget = children[z]->FindWidget( ID );
         if ( widget != NULL )
             return widget;
    }

    return NULL;
}

pawsWidget* pawsWidget::FindWidgetXMLBinding( const char* xmlbinding )
{
    csString temp(xmlbinding);
    if ( (this->xmlbinding == temp) || 
         (this->xmlbinding == "" && name == temp ))
        return this;

    for ( int z = 0; z < children.Length(); z++ )
    {
         pawsWidget* widget = children[z]->FindWidgetXMLBinding( xmlbinding );
         if ( widget != NULL )
             return widget;
    }

    Warning2( LOG_PAWS, "Could not locate widget with XML binding %s", xmlbinding );           
    return NULL;
}

void pawsWidget::DrawBackground()
{
	// Check for invalid clipping.
	if ((clipRect.ymin > clipRect.ymax) || (clipRect.xmin > clipRect.xmax))
		return;

    if ( bgColour != -1 )
    {        
        graphics2D->DrawBox( screenFrame.xmin,
                             screenFrame.ymin,
                             screenFrame.Width(), 
                             screenFrame.Height(),
                             bgColour );
    }

    if ( bgImage )
    {
        bool focus = hasFocus;
        // If my children are focused, I should fade like I am focused
        if (!focus)
        {
            for ( int x = children.Length()-1; x >= 0 && !focus ; x-- )
                focus = children[x]->HasFocus();
        }

 		// if the fading feature for this widget is enabled
		if (alpha && fade) 
		{
			// if the widget hasn't got the focus
			if (!focus)	
			{
				// and the mouse is inside the widget
				if (hasMouseFocus)
					// slowly become opaque
					fadeVal-=fadeSpeed;
				else
					// else slowly fade out
					fadeVal+=fadeSpeed;
			} else
				// when it has got the focus become/stay opaque
				fadeVal=0;

			if (fadeVal<0) fadeVal=0;
			if (fadeVal>100) fadeVal=100;

			bgImage->SetAlpha((int)(alphaMin + (alpha-alphaMin) * fadeVal * 0.010));
		}

		bgImage->Draw( screenFrame );  
    }
		
		if(fade && parent && parent->GetMaxAlpha() >= 0)
		{
			fadeVal = parent->GetFadeVal();
			alpha = parent->GetMaxAlpha();
			alphaMin = parent->GetMinAlpha();
			if ( bgImage )bgImage->SetAlpha((int)(alphaMin + (alpha-alphaMin) * fadeVal * 0.010));
		}

    if ( border )
        border->Draw();
        
    if ( IsResizable() )
        graphics2D->DrawBox( ScreenFrame().xmax-8, ScreenFrame().ymax-8, 8, 8, graphics2D->FindRGB( 255,255,255 ) );        
}

void pawsWidget::Draw()
{
    ClipToParent();
    DrawBackground();
    DrawChildren();
}


void pawsWidget::DrawWidgetText(const char *text, int x, int y, int style)
{
    csRef<iFont> font = GetFont();
    if (style==-1)
        style = GetFontStyle();

    if (style & FONT_DROPSHADOW)
        graphics2D->Write( font, x+2, y+2, GetFontShadowColour(), -1, text );

    graphics2D->Write( font, x, y, GetFontColour(), -1, text);    
}


void pawsWidget::DrawToolTip(int x, int y)
{	
	if (toolTip.Length() == 0)
		return;

	int width=0;
    int height=0;
	GetFont()->GetDimensions( toolTip , width, height );

	// Draw above the cursor for the time being
	y = y - height - 2;

	csRect rect(x+2, y+1, x+width+2, y+height+1);

	graphics2D->DrawBox( x,
			y,
			width + 4, 
			height + 2,
			graphics2D->FindRGB(0, 0, 0));

	graphics2D->Write( GetFont(), 
        x+2,
        y+1,
        graphics2D->FindRGB(255, 255, 255),
        graphics2D->FindRGB(0, 0, 0),
        toolTip);
}


void pawsWidget::DrawChildren()
{    
    for ( int x = children.Length()-1; x >= 0  ; x-- )
    {
        if ( children[x]->IsVisible() )
        {
            children[x]->Draw();
        }
    }
}

int pawsWidget::CalcChildPosition(pawsWidget * child)
{
    int pos;
    
    pos = 0;    
    if ( ! child->IsAlwaysOnTop() )
    {
        for ( int x = 0; x < children.Length(); x++ )
            if ( children[x]->IsAlwaysOnTop() )
                pos++;
    }
    return pos;
}
    
void pawsWidget::BringToTop( pawsWidget* widget )
{
    for ( int x = 0; x < children.Length(); x++ )
    {
        if ( children[x] == widget )
        {
            children.DeleteIndex(x);
            children.Insert(CalcChildPosition(widget), widget);
            break;
        }
    }

    if ( parent )
        parent->BringToTop( this );
}

void pawsWidget::SetTitle(const char *title)
{
	if (border!=NULL)
	{
		border->SetTitle(title);
		border->Draw();
	};
}


void pawsWidget::SaveSettings()
{
    csRef<iConfigManager> cfgMgr;
    cfgMgr = CS_QUERY_REGISTRY( windowManager->GetObjectRegistry(), iConfigManager );

    csRef<iConfigFile> config = cfgMgr->LookupDomain("/this/planeshift.cfg");        

    char buffer[512];

    sprintf(buffer, "PlaneShift.GUI.%s.MinTransparancy", GetName() );
    config->SetInt(buffer, alphaMin );

    sprintf(buffer, "PlaneShift.GUI.%s.MaxTransparancy", GetName() );
    config->SetInt(buffer, alpha );

	sprintf(buffer, "PlaneShift.GUI.%s.FadeSpeed", GetName() );
	config->SetFloat( buffer, fadeSpeed );

    sprintf(buffer, "PlaneShift.GUI.%s.Fade", GetName() );
    config->SetBool( buffer, fade );

    config->Save();
}


void pawsWidget::LoadSettings()
{
    csRef<iConfigManager> cfgMgr;
    cfgMgr = CS_QUERY_REGISTRY( windowManager->GetObjectRegistry(), iConfigManager );
  
    char buffer[512];

    csRef<iConfigFile> config = cfgMgr->LookupDomain("/this/planeshift.cfg");
 
    sprintf(buffer, "PlaneShift.GUI.%s.MinTransparancy", GetName() );
    alphaMin = config->GetInt (buffer, alphaMin );

    sprintf(buffer, "PlaneShift.GUI.%s.MaxTransparancy", GetName() );
    alpha = config->GetInt (buffer, alpha );

    sprintf(buffer, "PlaneShift.GUI.%s.Fade", GetName() );
    fade = config->GetBool (buffer, fade );

    sprintf(buffer, "PlaneShift.GUI.%s.FadeSpeed", GetName() );
    fadeSpeed = config->GetFloat (buffer, fadeSpeed );
}


void pawsWidget::CreateWidgetConfigWindow()
{
    psString title = name + "settings";
    pawsWidget * widget = windowManager->GetMainWidget()->FindWidget(title);
    
    // there can only be one config window active for each widget
	if ( widget ==NULL)
	{
		WidgetConfigWindow* configWindow = new WidgetConfigWindow(windowManager);
		configWindow->LoadFromFile("data/widgetconfigwindow.xml");
		configWindow->SetConfigurableWidget(this);
		configWindow->SetName(title);
		configWindow->Show();
		windowManager->GetMainWidget()->AddChild(configWindow);	
    }
    else
    {
        widget->Show();
    }
}



bool pawsWidget::CreateContextMenu()
{
    csPoint mousePos;
    
    mousePos = windowManager->GetMouse()->GetPosition();
    
    contextMenu = new pawsMenu(windowManager);
    if ( ! contextMenu->LoadFromFile(contextMenuFile) )
    {
        delete contextMenu;
        contextMenu = NULL;
        return false;
    }
    contextMenu->SetNotify(this);
    windowManager->GetMainWidget()->AddChild(contextMenu);
    contextMenu->CenterTo(mousePos.x, mousePos.y);
    contextMenu->MakeFullyVisible();
    contextMenu->Show();
    contextMenu->SetAlwaysOnTop(true);

	// add an 'settings' option to the menu if 'configurable'
	// is set to 'yes'
	if (configurable)
	{
		// add menu item "ConfigWindow"
		pawsMenuItem *item = new pawsMenuItem(windowManager);
		item->MakeFullyVisible();
		item->Show();
		item->SetAlwaysOnTop(true);
		item->SetColour(0xffffff);
		item->SetLabel("Settings");
		item->SetSizes(-1, 1, 1);
		item->SetName("Settings");

		// assign "configure"-action to this menu item
		pawsMenuAction action;
		action.name="configure";
		item->SetAction(action);
	//	item->SetConfigurableWidget(this);

		// add item to menu
		contextMenu->AddItem(item);
	}

    return true;
}

int pawsWidget::ResizeFlags( int x, int y )
{
    csRect frame = ScreenFrame();
    csRect hotButton( 0,0,0,0 );
    int flag = 0;
    
    hotButton= csRect( frame.xmax-8, 
                       frame.ymax-8, 
                       frame.xmax,
                       frame.ymax );
        
    if ( hotButton.Contains(x,y) )
    {
        flag |= RESIZE_BOTTOM;    
        flag |= RESIZE_RIGHT;
    }
    return flag; 
}

bool pawsWidget::OnMouseEnter()
{
    if ( parent )
        return parent->OnChildMouseEnter( this );
    else return true;        
}

bool pawsWidget::OnChildMouseEnter( pawsWidget* widget )
{
    return true;        
}

bool pawsWidget::OnMouseExit()
{
    if ( parent )
        return parent->OnChildMouseExit( this );
    else return true;        
}

bool pawsWidget::OnChildMouseExit( pawsWidget* child )
{
    return true;        
}


bool pawsWidget::OnMouseDown( int button, int modifiers, int x, int y )
{        
#ifdef PAWS_CONSTRUCTION
    if ( modifiers == 1 )
    {
        SetMovable(true);
        isResizable = true;
    }
    if ( modifiers == 2 )
    {
        printf("-----------WIDGET %s DUMP---------\n", name.GetData() );
        printf("Position: (%d,%d)\n", ScreenFrame().xmin, ScreenFrame().ymin);
        printf("Size    : (%d,%d)\n", ScreenFrame().Width(), ScreenFrame().Height());
    }                
#endif

    if (button == 1)
    {
        if ( isResizable  )
        {
            // Check to see if the mouse was in the resizing flags
            int flags = ResizeFlags(x,y) ;
            
            if ( flags )
            {
                windowManager->ResizingWidget( this, flags );
                    return true;
            }
        }


        if ( movable )
        {
            windowManager->MovingWidget( this );
            return true;
        }
    }
    else if (button == 2)
    {
        if ((contextMenu == NULL) && (contextMenuFile != ""))
            CreateContextMenu();
		else if ((contextMenuFile=="") && (configurable))
		{
			CreateWidgetConfigWindow();
		}
    }

    if ( parent )
        return parent->OnMouseDown( button, modifiers, x, y );

    return false;
}

bool pawsWidget::OnMenuAction ( pawsWidget * widget, const pawsMenuAction & action )
{
    if (action.name == MENU_DESTROY_ACTION_NAME)
    {       
        windowManager->GetMainWidget()->DeleteChild(widget);
        if (widget == contextMenu)
            contextMenu = NULL;
        return true;
    }

    if (action.name=="configure")
        CreateWidgetConfigWindow();

    return false;
}

csRect pawsWidget::ScreenFrame()
{ 
    return screenFrame; 
}


bool pawsWidget::OnMouseUp( int button, int modifiers, int x, int y )
{
    if ( parent )
        return parent->OnMouseUp( button, modifiers, x, y );
    else
        return false;
}

bool pawsWidget::OnKeyDown( int keyCode, int key, int modifiers )
{
    if (keyCode==CSKEY_TAB)
    {
        int z;

        for (z = 0; z < children.Length(); z++ )
        {
            if (children[z]->HasFocus())
            {
                break;
            }
        }

        // One of the widget's children is focused, so pass it to the next child
        if (z < children.Length())
        {
            // Focus the next child
            if (z + 1 < children.Length())
                z++;
            else
                z = 0;

            // Check if widget CAN be focused
            for(;!(children[z]->OnGainFocus(false));)
            {
                if(z + 1 < children.Length())
                    z++;
                else
                    z = 0;
            }
            windowManager->SetCurrentFocusedWidget(children[z]);
        }
    }
    
    if ( parent ) 
        return parent->OnKeyDown( keyCode, key, modifiers );

    return false;
}

void pawsWidget::MoveDelta( int deltaX, int deltaY )
{
    screenFrame.xmin += deltaX;
    screenFrame.ymin += deltaY;
    screenFrame.xmax += deltaX;
    screenFrame.ymax += deltaY;

    for ( int x = children.Length()-1; x >= 0  ; x-- )
    {
        children[x]->MoveDelta( deltaX, deltaY );
    }
}         


void pawsWidget::MoveTo( int x, int y )
{
    int deltaX = x - screenFrame.xmin;
    int deltaY = y - screenFrame.ymin;

    int width = screenFrame.Width();
    int height = screenFrame.Height();

    screenFrame.xmin = x;
    screenFrame.ymin = y;
    screenFrame.SetSize( width, height );

    for ( int z = children.Length()-1; z >= 0  ; z-- )
    {
        children[z]->MoveDelta(deltaX, deltaY );
    }
    
    for ( int z = 0; z < children.Length(); z++ )
    {    
        children[z]->RecalcScreenPositions();
    } 
}

void pawsWidget::CenterTo( int x, int y )
{
    MoveTo(x - screenFrame.Width()/2, y - screenFrame.Height()/2);
}

csRect pawsWidget::LoadPositions()
{    
    csRef<iConfigManager> cfgMgr;
    cfgMgr = CS_QUERY_REGISTRY( windowManager->GetObjectRegistry(), iConfigManager );
  
    char buffer[512];

    csRef<iConfigFile> config = cfgMgr->LookupDomain("/this/planeshift.cfg");
 
    sprintf(buffer, "PlaneShift.GUI.%s.PosX", GetName() );
    int winPosX = config->GetInt (buffer, defaultFrame.xmin );

    sprintf(buffer, "PlaneShift.GUI.%s.PosY", GetName() );
    int winPosY = config->GetInt (buffer, defaultFrame.ymin );

    //don't do this if widget is not resizable
    int winWidth = 0;
    int winHeight = 0;
    if( IsResizable() )
    {
        sprintf(buffer, "PlaneShift.GUI.%s.Width", GetName() );
        winWidth = config->GetInt (buffer, defaultFrame.Width() );

        sprintf(buffer, "PlaneShift.GUI.%s.Height", GetName() );
        winHeight = config->GetInt (buffer, defaultFrame.Height() );
    }

    int screenWidth  = graphics2D->GetWidth();
    int screenHeight = graphics2D->GetHeight();

    winPosX = GetActualWidth( winPosX );
    winPosY = GetActualHeight( winPosY );

    if( IsResizable() )
    {
        winWidth = GetActualWidth( winWidth );
        winHeight = GetActualHeight( winHeight );
    
        if ( winWidth < min_width ) winWidth = min_width;
        if ( winHeight < min_height ) winHeight = min_height;
    }

    if ( winPosX < 0 ) winPosX = 0;
    if ( winPosY < 0 ) winPosY = 0;

    if ( winPosX + winWidth  > screenWidth  ) winPosX = screenWidth  - winWidth;
    if ( winPosY + winHeight > screenHeight ) winPosY = screenHeight - winHeight;

    csRect newPosition = csRect(winPosX,winPosY, winPosX+winWidth, winPosY+winHeight);
    return newPosition;
}

void pawsWidget::PerformAction( const char* action )
{
    csString temp(action);
    if ( temp == "togglevisibility" )
    {
        if ( IsVisible() ) 
        {
            Hide(); 
        }
        else 
        {
            Show(); 
        }
        return;
    }
}

void pawsWidget::SavePosition()
{
    csRef<iConfigManager> cfgMgr;
    cfgMgr = CS_QUERY_REGISTRY( windowManager->GetObjectRegistry(), iConfigManager );

    csRef<iConfigFile> config = cfgMgr->LookupDomain("/this/planeshift.cfg");        

    char buffer[512];

    int desktopWidth  = graphics2D->GetWidth(); 
    int desktopHeight = graphics2D->GetHeight();

    float ratioX = (float) 800.0f / (float)desktopWidth;
    float ratioY = (float) 600.0f / (float)desktopHeight;

    sprintf(buffer, "PlaneShift.GUI.%s.PosX", GetName() );
    config->SetInt( buffer, RX(screenFrame.xmin) );

    sprintf(buffer, "PlaneShift.GUI.%s.PosY", GetName() );
    config->SetInt( buffer, RY(screenFrame.ymin) );

    if ( IsResizable() )
    {
        sprintf(buffer, "PlaneShift.GUI.%s.Width", GetName() );
        config->SetInt( buffer, RX(screenFrame.Width()) );

        sprintf(buffer, "PlaneShift.GUI.%s.Height", GetName() );
        config->SetInt( buffer, RY(screenFrame.Height()) );
    }

    config->Save();
}

void pawsWidget::SetSize( int newWidth, int newHeight )
{   
    screenFrame.SetSize( newWidth, newHeight );
    OnResize();
    
    for ( int x = children.Length()-1; x >= 0  ; x-- )
    {
        children[x]->Resize();
    }    
}

void pawsWidget::StopResize()
{
    for ( int x = 0; x < children.Length(); x++ )
    {
        children[x]->StopResize();
    }
}

void pawsWidget::Resize( int deltaX, int deltaY, int flags )
{
    csRect newFrame = screenFrame;
    if ( flags & RESIZE_LEFT)
    {
        newFrame.xmin += deltaX;
    }
    if ( flags & RESIZE_RIGHT )
    {
        newFrame.xmax += deltaX;
    }
    if ( flags & RESIZE_TOP )
    {
        newFrame.ymin += deltaY;
    }
    if ( flags & RESIZE_BOTTOM )
    {
        newFrame.ymax += deltaY;
    }


    /// prevent the widget from going below its min height and width
    if(newFrame.xmax - newFrame.xmin < min_width) newFrame.xmax = newFrame.xmin + min_width;
    if(newFrame.ymax - newFrame.ymin < min_height) newFrame.ymax = newFrame.ymin + min_height;

    /// prevent the widget from going beyond its max size
    if(newFrame.xmax - newFrame.xmin > max_width) newFrame.xmax = max_width + newFrame.xmin;
    if(newFrame.ymax - newFrame.ymin > max_height) newFrame.ymax = max_height + newFrame.ymin;

    screenFrame = newFrame;
    
    OnResize();

    for ( int x = children.Length()-1; x >= 0  ; x-- )
    {
        children[x]->Resize();
    }    
}  


void pawsWidget::Resize()
{
    if ( attachFlags & ATTACH_LEFT )
    {
        int width = screenFrame.Width();

        screenFrame.xmin = parent->ScreenFrame().xmin + defaultFrame.xmin;    
        screenFrame.SetSize( width, screenFrame.Height() );

        // Left-Right Attachment
        if ( attachFlags & ATTACH_RIGHT )
        {
            int newX = parent->DefaultFrame().Width() - (defaultFrame.xmin + defaultFrame.Width());        
            screenFrame.xmax = parent->ScreenFrame().xmax - newX;
        }   

        // Left-Right Proportional Attachment
        if ( attachFlags & PROPORTIONAL_RIGHT )
        {
            int z = (parent->ScreenFrame().Width() * ( parent->DefaultFrame().Width() - defaultFrame.xmax ) ) /
                        parent->DefaultFrame().Width();

            int newMax = parent->ScreenFrame().xmax - z;
            screenFrame.xmax = newMax;
        }
    }                 

    if ( attachFlags & ATTACH_RIGHT )
    {
        //Attach Right Only
        if ( !(attachFlags & ATTACH_LEFT) )
        {            
            int width = screenFrame.Width();
            screenFrame.xmin = parent->ScreenFrame().xmax - ( parent->DefaultFrame().Width() - defaultFrame.xmax + width );
            screenFrame.SetSize( width, screenFrame.Height() );
        }   

        //Attach Right-Left Proportional               
        if ( attachFlags & PROPORTIONAL_LEFT )
        {            
            int z = (parent->ScreenFrame().Width() *  defaultFrame.xmin ) /
                        parent->DefaultFrame().Width();

            int newMin = parent->ScreenFrame().xmin + z;
            screenFrame.xmin = newMin;            
        }
    }

    if ( attachFlags & PROPORTIONAL_LEFT )
    {
        //Attach Proportional Left Only
        if ( ! (attachFlags & ATTACH_RIGHT) && !(attachFlags & PROPORTIONAL_RIGHT) )
        {
            int z = (parent->ScreenFrame().Width() * defaultFrame.xmin ) /
                        parent->DefaultFrame().Width();

            int newMin = parent->ScreenFrame().xmin + z;
            int width = screenFrame.Width();
            screenFrame.xmin = newMin;      
            screenFrame.SetSize( width, screenFrame.Height() );      
        }                                        
    }

    if ( attachFlags & PROPORTIONAL_RIGHT )
    {       
        //Attach Proportional RIGHT Only
        if ( ! (attachFlags & ATTACH_LEFT) && !(attachFlags & PROPORTIONAL_LEFT) )
        {

            int z = (parent->ScreenFrame().Width() * ( parent->DefaultFrame().Width() - defaultFrame.xmax ) ) /
                        parent->DefaultFrame().Width();

            int newMin = parent->ScreenFrame().xmax - z - defaultFrame.Width();
            int width = screenFrame.Width();
            screenFrame.xmin = newMin;      
            screenFrame.SetSize( width, screenFrame.Height() );      
        }                                        

        //Attach PropLeft-PropRight
        if ( attachFlags & PROPORTIONAL_LEFT )
        {
            int z = (parent->ScreenFrame().Width() * ( parent->DefaultFrame().Width() - defaultFrame.xmax ) ) /
                        parent->DefaultFrame().Width();

            int newMax = parent->ScreenFrame().xmax - z;
            screenFrame.xmax = newMax;

            z = (parent->ScreenFrame().Width() *  defaultFrame.xmin ) /
                        parent->DefaultFrame().Width();

            int newMin = parent->ScreenFrame().xmin + z;
            screenFrame.xmin = newMin;            
        }
    }






    if ( attachFlags & ATTACH_TOP )
    {
        int height = defaultFrame.Height();
                  
        screenFrame.ymin = parent->ScreenFrame().ymin + defaultFrame.ymin; 
        screenFrame.SetSize( screenFrame.Width(), height );

        //Attach TOP-BOTTOM
        if ( attachFlags & ATTACH_BOTTOM )
        {            
            int newY = parent->DefaultFrame().Height() - (defaultFrame.ymin + defaultFrame.Height());        
            screenFrame.ymax = parent->ScreenFrame().ymax - newY;
        }   

        //Attach TOP-BOTTOM Proportional
        if ( attachFlags & PROPORTIONAL_BOTTOM )
        {
            int z = (parent->ScreenFrame().Height() * ( parent->DefaultFrame().Height() - defaultFrame.ymax ) ) /
                        parent->DefaultFrame().Height();

            int newMax = parent->ScreenFrame().ymax - z;
            screenFrame.ymax = newMax;
        }
    }                 

    if ( attachFlags & ATTACH_BOTTOM )
    {
        // Attach BOTTOM Only
        if ( !(attachFlags & ATTACH_TOP) )
        {
         
            int height = screenFrame.Height();
            screenFrame.ymin = parent->ScreenFrame().ymax - ( parent->DefaultFrame().Height() - defaultFrame.ymax + height );
            screenFrame.SetSize( screenFrame.Width(), height );
        } 
          
        //Attach BOTTOM-TOP Proportional
        if ( attachFlags & PROPORTIONAL_TOP )
        {            
            int z = (parent->ScreenFrame().Height() *  defaultFrame.ymin ) /
                        parent->DefaultFrame().Height();

            int newMin = parent->ScreenFrame().ymin + z;
            screenFrame.ymin = newMin;            
        }
    }

    // Attach Proportional TOP Only
    if ( attachFlags & PROPORTIONAL_TOP )
    {
        if ( ! (attachFlags & ATTACH_BOTTOM) && !(attachFlags & PROPORTIONAL_BOTTOM) )
        {
            int z = (parent->ScreenFrame().Height() * defaultFrame.ymin ) /
                        parent->DefaultFrame().Height();

            int newMin = parent->ScreenFrame().ymin + z;
            int height = screenFrame.Height();
            screenFrame.ymin = newMin;      
            screenFrame.SetSize( screenFrame.Width(), height );      
        }                                        
    }


    if ( attachFlags & PROPORTIONAL_BOTTOM )
    {
        // Attach Proportional BOTTOM Only
        if ( ! (attachFlags & ATTACH_TOP) && !(attachFlags & PROPORTIONAL_TOP) )
        {
         

            int z = (parent->ScreenFrame().Height() * ( parent->DefaultFrame().Height() - defaultFrame.ymax ) ) /
                        parent->DefaultFrame().Height();

            int newMin = parent->ScreenFrame().ymax - z - defaultFrame.Height();
            int height = screenFrame.Height();
            screenFrame.ymin = newMin;      
            screenFrame.SetSize( screenFrame.Width(), height );      
        }                                        

         // Attach PropTOP-PropBOTTOM
        if ( attachFlags & PROPORTIONAL_TOP )
        {            
            int z = (parent->ScreenFrame().Height() * ( parent->DefaultFrame().Height() - defaultFrame.ymax ) ) /
                        parent->DefaultFrame().Height();

            int newMax = parent->ScreenFrame().ymax - z;
            screenFrame.ymax = newMax;

            z = (parent->ScreenFrame().Height() *  defaultFrame.ymin ) /
                        parent->DefaultFrame().Height();

            int newMin = parent->ScreenFrame().ymin + z;
            screenFrame.ymin = newMin;            
        }
    }
    OnResize();    
    for ( int x = children.Length()-1; x >= 0  ; x-- )
    {    
        children[x]->Resize();
    }    
}       

void pawsWidget::OnResize()
{
    SetCloseButtonPos();
}

void pawsWidget::SetModalState( bool isModal )
{
    if ( isModal )
        windowManager->SetModalWidget( this );
    else
        windowManager->SetModalWidget( 0 );
}



int pawsWidget::GetAttachFlag( const char* flag )
{    
    csString str( flag );

    if ( str == "ATTACH_RIGHT" ) return ATTACH_RIGHT;
    if ( str == "ATTACH_LEFT" )  return ATTACH_LEFT;
    if ( str == "ATTACH_BOTTOM" ) return ATTACH_BOTTOM;
    if ( str == "ATTACH_TOP" ) return ATTACH_TOP;

    if ( str == "PROPORTIONAL_LEFT" ) return PROPORTIONAL_LEFT;
    if ( str == "PROPORTIONAL_RIGHT" ) return PROPORTIONAL_RIGHT;  
    if ( str == "PROPORTIONAL_TOP" ) return  PROPORTIONAL_TOP;
    if ( str == "PROPORTIONAL_BOTTOM" ) return  PROPORTIONAL_BOTTOM;

    return 0;
}

void pawsWidget::MouseOver(bool value)
{
	hasMouseFocus = value;
}

csRef<iDocument> ParseFile(iObjectRegistry* object_reg, const csString & name);

bool pawsWidget::LoadFromFile(const csString & fileName)
{
    csRef<iDocument> doc;
    csRef<iDocumentNode> root, topNode, widgetNode;
        
    doc = ParseFile(windowManager->GetObjectRegistry(), 
                    windowManager->GetLocalization()->FindLocalizedFile(fileName));
    if (doc == NULL)
        return false;
        
    root = doc->GetRoot();
    if (root == NULL)
    {
        printf("No root in XML");
        return false;
    }
    topNode = root->GetNode("widget_description");
    if (topNode == NULL)
    {
        printf("No <widget_description> in XML");
        return false;
    }
    widgetNode = topNode->GetNode("widget");
    if (widgetNode == NULL)
    {
        printf("No <widget> in <widget_description>");
        return false;
    }
    
    if ( ! Load(widgetNode) )
        return false;
    
    return true;
}

void pawsWidget::MakeFullyVisible()
{
    csRect parentFrame;
    int targetX, targetY;
    
    if (parent == NULL)
        return;
    
    parentFrame = parent->ScreenFrame();
    
    if (screenFrame.xmin < parentFrame.xmin)
        targetX = parentFrame.xmin;
    else if (screenFrame.xmax > parentFrame.xmax)
        targetX = parentFrame.xmax - screenFrame.Width() + 1;
    else
        targetX = screenFrame.xmin;
    if (screenFrame.ymin < parentFrame.ymin)
        targetY = parentFrame.ymin;
    else if (screenFrame.ymax > parentFrame.ymax)
        targetY = parentFrame.ymax - screenFrame.Height() + 1;
    else
        targetY = screenFrame.ymin;
    
    if ((targetX != screenFrame.xmin) || (targetY != screenFrame.ymin))
        MoveTo(targetX, targetY);
}

void pawsWidget::ClipToParent()
{    
    if ( !parent )
    {
        graphics2D->SetClipRect( 0,0, graphics2D->GetWidth(), graphics2D->GetHeight());
        clipRect = csRect(0,0, graphics2D->GetWidth(), graphics2D->GetHeight());
    }
    else
    {
        if ( parent->ClipRect().Width() == 0 || parent->ClipRect().Height() == 0 )
        {
            clipRect = screenFrame;
        }
        else
        {
            clipRect.xmin = MAX( parent->ClipRect().xmin, ScreenFrame().xmin );
            clipRect.ymin = MAX( parent->ClipRect().ymin, ScreenFrame().ymin );
            clipRect.xmax = MIN( parent->ClipRect().xmax, ScreenFrame().xmax );
            clipRect.ymax = MIN( parent->ClipRect().ymax, ScreenFrame().ymax );
		}            
        
        if ( border && border->GetTitleImage())
            clipRect = border->GetRect();
                
        graphics2D->SetClipRect( ClipRect().xmin, ClipRect().ymin, 
                                 ClipRect().xmax, ClipRect().ymax );
    }

}

void pawsWidget::LoadBorderColours( iDocumentNode* node )
{
    csRef<iDocumentNodeIterator> iter = node->GetNodes();
    
    int index = 0;
    while ( index < 5 && iter->HasNext() )
    {
        csRef<iDocumentNode> colour = iter->Next();
        
        int r = colour->GetAttributeValueAsInt( "r" );            
        int g = colour->GetAttributeValueAsInt( "g" );            
        int b = colour->GetAttributeValueAsInt( "b" );            
        
        int col = graphics2D->FindRGB( r, g, b );
        borderColours[index++] = col;
    }    
    hasBorderColours = true;
}

int pawsWidget::GetBorderColour(int which)
{
    if (hasBorderColours)
        return borderColours[which];
    else if (parent)
        return parent->GetBorderColour(which);
    else
        return windowManager->GetPrefs()->GetBorderColour(which);
}

void pawsWidget::SetFont( const char* Name , int Size)
{
    if (fontName)
	{
        myFont = graphics2D->GetFontServer()->LoadFont( fontName, Size );
        fontSize = Size;
        fontName = Name;
	}
    else
    {
        myFont = NULL;
    }
}

void pawsWidget::SetColour( int newColour )
{
    if ( newColour != -1 )
    {
        defaultFontColour = newColour;
    }
    else
    {
        defaultFontColour = windowManager->GetPrefs()->GetDefaultFontColour();
    }
}

void pawsWidget::ChangeFontSize(int newSize )
{
    fontSize = newSize;
    myFont = graphics2D->GetFontServer()->LoadFont( fontName, fontSize );
}

iFont *pawsWidget::GetFont()
{
    csRef<iFont>  font;

    if (myFont)
        font = myFont;
    else if (parent)
        font = parent->GetFont();
    else
        font = windowManager->GetPrefs()->GetDefaultFont();

    return font;
}

int pawsWidget::GetFontColour()
{
    if (myFont)
        return defaultFontColour;
    else if (parent)
        return parent->GetFontColour();
    else
        return windowManager->GetPrefs()->GetDefaultFontColour();
}

int pawsWidget::GetFontShadowColour()
{
    if (myFont)
        return defaultFontShadowColour;
    else if (parent)
        return parent->GetFontShadowColour();
    else
        return graphics2D->FindRGB(0,0,0);
}

int pawsWidget::GetFontSize()
{
    if (myFont)
        return fontSize;
    else if (parent)
        return parent->GetFontSize();
    else
        return DEFAULT_FONT_SIZE;
}

int pawsWidget::GetFontStyle()
{
    if (myFont)
        return fontStyle;
    else if (parent)
        return parent->GetFontStyle();
    else
        return DFFAULT_FONT_STYLE;   
}

bool pawsWidget::SelfPopulateXML( const char *xmlstr )
{
    csRef<iDocumentSystem> xml = csPtr<iDocumentSystem>(new csTinyDocumentSystem);
    
    csRef<iDocument> doc= xml->CreateDocument();
    const char *error = doc->Parse( xmlstr );
    if (error)
    {
        printf(error);
        return false;
    }
    csRef<iDocumentNode> root    = doc->GetRoot();
    csRef<iDocumentNode> topNode = root->GetNode(xmlbinding!=""?xmlbinding:name);
    if (topNode)
    {
        return SelfPopulate(topNode);
    }
    else
    {
        printf("SelfPopulate failed because tag <%s> could not be found in xmlstr '%s'.\n",name.GetData(),xmlstr);
        return false;
    }
}

bool pawsWidget::SelfPopulate( iDocumentNode *node)
{
    // No self population here since generic widgets have no data

    // Now generically check to populate children if any are here
    if (!children.Length())
        return true;

    csRef<iDocumentNodeIterator> iter = node->GetNodes();

    int count = 0;
    while ( iter->HasNext() )
    {
        csRef<iDocumentNode> node = iter->Next();

        if ( node->GetType() != CS_NODE_ELEMENT )
            continue;

        pawsWidget *nodewidget = FindWidgetXMLBinding(node->GetValue());
        if (nodewidget)
        {
            if (!nodewidget->SelfPopulate(node))
                return false;
        }
    }
    return true;
}

bool pawsWidget::OnChange(pawsWidget * widget)
{
    if (parent != NULL)
        return parent->OnChange(widget);
    else
        return false;
}

void pawsWidget::Dump(csString tab)
{
    printf("%s<widget name=\"%s\">\n",tab.GetData(),name.GetData());
    for ( int x = 0; x < children.Length(); x++ )
    {
        children[x]->Dump(tab+"\t");
    }
    printf("%s</widget>\n",tab.GetData());
}

//------------------------------------------------------------------------

pawsWidgetFactory::pawsWidgetFactory( PawsManager* manager )
{
    mainManager = manager;
}

pawsWidgetFactory::~pawsWidgetFactory()
{
}

void pawsWidgetFactory::Register( const char* name )
{
    factoryName = name;

    mainManager->RegisterWidgetFactory( this );
}

pawsWidget* pawsWidgetFactory::Create()
{
    return NULL;
}

void pawsWidget::SetCloseButtonPos()
{
    pawsButton * button = dynamic_cast <pawsButton*> (FindWidget(GetCloseName()));
    if (button != NULL)
        button->SetRelativeFrame( defaultFrame.Width() - 20, - 16, 16, 16 );
}
    
bool pawsWidget::SetTitle( const char* text, const char* image, const char* align, const char* close_button )
{
    if( !border )
    {
        return false;
    }        
    else
    {
        border->SetTitle( text );
        psImageDescription* imageDesc = windowManager->GetTextureManager()->GetImageDescription( image );        
        pawsImage *titleImage = new pawsImage( windowManager->GetObjectRegistry() );
        titleImage->Description( imageDesc );
        border->SetTitleImage(titleImage);

        int alignValue = ALIGN_LEFT;
        if (align)
        {
            if (!strcmp(align,"center"))
                alignValue = ALIGN_CENTER;
            else if (!strcmp(align,"right"))
                alignValue = ALIGN_RIGHT;
        }
        border->SetTitleAlign(alignValue);
        
        
        if ( close_button && (strcmp(close_button,"yes") == 0) )
        {
            pawsButton* button = (pawsButton*)WindowManager()->CreateWidget("pawsButton");
            button->SetParent( this );
            button->SetBackground( "quit" );
            csString name( GetName() );
            name.Append( "_close" );
            button->SetName( name );
            SetCloseButtonPos();
            
            int flags = ATTACH_TOP | ATTACH_RIGHT;
            button->SetAttachFlags( flags );
            AddChild( button );
        }   
    }  
    
    return true;                                      
}

csString pawsWidget::GetPathInWidgetTree()
{
    csString path, widget;
    
    widget =  "n=";
    widget += name;
    widget +=" f=";
    widget += factory;
    
    if (parent != NULL)
    {
        csString path;
        path =  parent->GetPathInWidgetTree();
        path += " ---> ";
        path += widget;
        return path;
    }
    else
        return widget;
}
