/*
 * pawswidget.h - Author: Andrew Craig
 *
 * Copyright (C) 2003 PlaneShift Team (info@planeshift.it,
 * http://www.planeshift.it)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation (version 2 of the License)
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */
// pawswidget.h: interface for the pawsWidget class.
//
//////////////////////////////////////////////////////////////////////

#ifndef PAWS_WIDGET_HEADER
#define PAWS_WIDGET_HEADER

#include <ivideo/graph2d.h>

#include <csutil/array.h>
#include <csutil/csstring.h>

#include <csgeom/cspoint.h>
#include <csgeom/csrectrg.h>
#include <csgeom/vector2.h>

#include <iutil/document.h>

#define ATTACH_LEFT     2
#define ATTACH_RIGHT    4
#define ATTACH_BOTTOM   8
#define ATTACH_TOP      16
#define PROPORTIONAL_LEFT   32 
#define PROPORTIONAL_RIGHT  64 
#define PROPORTIONAL_TOP    128 
#define PROPORTIONAL_BOTTOM 256 

#define SCROLL_UP     -1000
#define SCROLL_DOWN   -2000
#define SCROLL_THUMB  -3000
#define SCROLL_AUTO   -4000
#define SCROLL_SET    -5000

#define DEFAULT_MIN_HEIGHT  10
#define DEFAULT_MIN_WIDTH   10


#define RESIZE_LEFT    2 
#define RESIZE_RIGHT   4 
#define RESIZE_TOP     8
#define RESIZE_BOTTOM 16

enum 
{
    ALIGN_LEFT,
    ALIGN_CENTER,
    ALIGN_RIGHT
};

#define BORDER_INNER_BEVEL 0x1
#define BORDER_OUTER_BEVEL 0x2
#define BORDER_REVERSED    0x4
#define BORDER_BUMP        (BORDER_OUTER_BEVEL)
#define BORDER_RAISED      (BORDER_INNER_BEVEL|BORDER_OUTER_BEVEL)
#define BORDER_SUNKEN      (BORDER_REVERSED|BORDER_OUTER_BEVEL)
#define BORDER_ETCHED      (BORDER_REVERSED|BORDER_RAISED)


#define TEXT_DROPSHADOW 1


class PawsManager;
class pawsImage;
class pawsBorder;
class pawsScrollBar;
class pawsListBox;
class pawsMenu;
class pawsMenuAction;
class WidgetConfigWindow;

/** The main base widget that all other widgets should inherrit from.
 */
class pawsWidget  
{
public:
    pawsWidget( PawsManager* manager );

    virtual ~pawsWidget();

    /** Locate a widget that is at these screen coordindates. 
     * This will recurse through all the children as well looking
     * for the lowest widget that contains these coordinates. 
     *
     * @param x The x screen position
     * @param y The y screen position
     * @return NULL if this widget does not contain these coordinates. 
     *         this if it contains these coordinates and none of it's children
     *         do.
     */
    pawsWidget* WidgetAt( int x, int y );

    /** Does an action based on this string.
     */
    virtual void PerformAction( const char* action );

    virtual csRect ScreenFrame();
    virtual csRect DefaultFrame() { return defaultFrame; }

    bool IsVisible() { return visible; }
    void SetVisibility(bool visible) { if (visible) Show(); else Hide(); }
    virtual void Show();
    virtual void Hide();
    virtual void Close() { Hide(); }
    
    
    virtual bool CheckButtonPressed( int button, int modifiers, pawsWidget* pressedWidget );
                            

    /** Add a child widget to this widget.  
     * This widget is then responsible for deleting this child.
     * @param widget The child to add.
     */
    void AddChild( pawsWidget* widget );
    
    /** Removes the widget from list of children and destructs it
     */
    void DeleteChild( pawsWidget* widget );

    /** Removes the widget from list of children but does NOT destruct it
        OWNERSHIP goes to caller
     */
    void RemoveChild( pawsWidget* widget );

    /** Set the owner of this widget.
     * @param widget The onwner of this widget.
     */
    void SetParent( pawsWidget* widget );

    /** Find a child widget of this widget.
     * Recurses down through the children as well looking for widget.
     * @param name The name of the widget to look for.
     * @return A pointer to the widget if found. NULL otherwise. 
     */
    pawsWidget* FindWidget( const char* name );

    /** Find a child widget of this widget.
     * Recurses down through the children as well looking for widget.
     * @param id The id of the widget to look for.
     * @return A pointer to the widget if found. NULL otherwise. 
     */    
    pawsWidget* FindWidget( int id );

    /** Find a child widget of this widget with the given XML binding.
     * Recurses down through the children as well looking for widget.
     * @param xmlbinding The xmlbinding of the widget to look for.
     * @return A pointer to the widget if found. NULL otherwise. 
     */
    pawsWidget* FindWidgetXMLBinding( const char* xmlbinding );

    virtual const csString& GetXMLBinding(){ return xmlbinding; }
    virtual void SetXMLBinding(csString & xmlbinding) { this->xmlbinding = xmlbinding; }

    pawsWidget* GetParent() { return parent; }

    /** Load a widget based on its <widget></widget> tag.
     * This class will also recursivly load all the child widgets as well.
     */
    virtual bool Load( iDocumentNode* node );
    
    /**
     * Load standard widget attributes based on its <widget></widget> tag.
     */
    virtual bool LoadAttributes( iDocumentNode* node );

    /**
     * Load widget children based on subtags of its <widget></widget> tag.
     */
    virtual bool LoadChildren( iDocumentNode* node );

    /** Parses XML file 'fileName', finds first widget tag and Load()s itself from this tag.
     * 'fileName' is standard path of the XML file to load - see psLocalization::FindLocalizedFile()
     */
    bool LoadFromFile(const csString & fileName);

    /** Setup this widget.
    */
    virtual bool Setup( iDocumentNode* node ) { return true; }

    /** This is called after the widget and all of it's children have been created.
     * This can be useful for widgets that want to get pointers to some of it's 
     * children for quick access.
     */
    virtual bool PostSetup() { return true; }

    /**
     * This function parses the xml string and calls SelfPopulate
     * with the resulting DOM structure if valid.  This is NOT the same as Setup,
     * which is defining the widget structure.  This is for data.
     */
    virtual bool SelfPopulateXML( const char *xmlstr );
    /**
     * This function allows a widget to fill in its own contents from an xml node
     * supplied. And call the same function for all childs. User should overload
     * this function to populate diffrent types of widgets. And call this base
     * class function to populate childs.
     * This is NOT the same as Setup, which is defining height, width,
     * styles, etc.  This is data.
     */
    virtual bool SelfPopulate( iDocumentNode *node );

    virtual void Draw();
    virtual void DrawBackground();
    virtual void DrawChildren();

    bool IsResizable() { return isResizable; }

    void SetModalState( bool isModal );

    // Will return FALSE if widget cannot be focused.
    // If the widget cannot be focused, this function should be overridden.
    virtual bool OnGainFocus( bool notifyParent = true ) 
    { 
        hasFocus = true; 
        if ( hasFocus ) return true;
        if ( parent ) parent->OnGainFocus();
        return true;
    }

    virtual void OnLostFocus() 
    { 
        if ( !hasFocus ) return; 
        hasFocus = false; 
        if ( parent ) parent->OnLostFocus();
    }
    
    virtual void OnResize();
    
    /// Called once the mouse up is done after resizing a widget.
    virtual void StopResize();

    virtual int GetBorderColour(int which);
    
    /** Set the position of this widget relative to the parent.
     */
    virtual void SetRelativeFrame( int x, int y, int width, int height );
    virtual void SetRelativeFramePos( int x, int y );
    virtual void SetRelativeFrameSize( int width, int height );
    
    virtual void SetAttachFlags( int flags ) { attachFlags = flags; }

    virtual bool OnMouseDown( int button, int modifiers, int x, int y );
    virtual bool OnMouseUp( int button, int modifiers, int x, int y );
    
    /// Called whenever the mouse enters this widget.
    virtual bool OnMouseEnter();
    
    /// Called whenever the mouse leaves this widget.
    virtual bool OnMouseExit();
    
    /// Called when mouse enters a child widget. ( If child wants to inform parent )
    virtual bool OnChildMouseEnter( pawsWidget* child );
    
    /// Called when a mouse exits a child widget. ( If child wants to infor parent )
    virtual bool OnChildMouseExit( pawsWidget* child );
    
    virtual bool OnKeyDown( int keyCode, int keyChar, int modifiers );
    virtual bool HasFocus() { return hasFocus; }

    /** Move this widget up the z order to the top
     *     - if the widget is alwaysOnTop, it will be placed on 0. position
     *     - if the widget is not alwaysOnTop, it will be placed after the last alwaysOnTop widget
     * Will also recurse up through the parents to bring it up as well.
     */
    virtual void BringToTop( pawsWidget* widget );

    /** Get the name of this widget
    */
    const char* GetName()  { return name; }  
    void SetName( const char* newName ) 
    { 
        name.Replace( newName ); 
        closeName = name;
        closeName.Append("_close");
    }
    const char* GetCloseName() { return closeName; }
    
    PawsManager* WindowManager() { return windowManager; }   

    /** Move a widget by a delta amount.
     * This will recurse through all the children calling MoveDelta on each.
     * @param dx Delta x
     * @param dy Delta y
     */
    virtual void MoveDelta( int dx, int dy );

    /** Move this widget to a location.
     * Moves this location ( and all of its children ) to a new screen location.
     * @param x Screen coordinate X.
     * @param y Screen coordinage Y.
     */
    virtual void MoveTo( int x, int y );
    
    /** Move this widget so that its center is at given location.
     * @param x Screen coordinate X.
     * @param y Screen coordinage Y.
     */
    virtual void CenterTo( int x, int y );

    /** Resize a widget by a delta amount.
     * This will recurse through all the children calling MoveDelta on each.
     * @param dx Delta x
     * @param dy Delta y
     * @param flags  The resize direction.
     */    
    virtual void Resize( int dx, int dy, int flags );

    /** Set the size of a particlar widget.
     * This will also call Resize() all of the children widgets to 
     * adjust to the new parent size. 
     * @param newWidth The new width of this widget.
     * @param newHeight The new height of this widget.
     */
    virtual void SetSize( int newWidth, int newHeight );

    /** This will make this widget use a border.
     */
    virtual void UseBorder( const char* style = 0 );
    virtual pawsBorder* GetBorder() { return border; }
    virtual int GetBorderStyle() { return BORDER_BUMP; }    

    virtual void SetBackground( const char* imageName );
    csString GetBackground();
    virtual void SetBackgroundAlpha( int alphaValue );

    /** Resize a widget based on it's parent's size.
     */
    virtual void Resize();
    
    virtual bool OnButtonPressed( int button, int keyModifier, pawsWidget* widget ) 
    { 
        if ( parent ) return parent->OnButtonPressed( button, keyModifier, widget );
        return false; 
    }

    virtual bool OnButtonReleased( int button, pawsWidget* widget)
    {
        if ( parent ) return parent->OnButtonReleased( button, widget );
        return false;
    }

    virtual bool OnScroll( int scrollDirection, pawsScrollBar* widget )
    {
        if ( parent ) return parent->OnScroll( scrollDirection, widget );
        return false;
    }

    virtual bool OnSelected( pawsWidget* widget) { return false; }
    
    virtual bool OnMenuAction ( pawsWidget * widget, const pawsMenuAction & action );
    
    virtual bool OnChange(pawsWidget * widget);

    void SetID( int newID ) { id = newID; }
    int  GetID() { return id; }

    bool IsAlwaysOnTop() { return alwaysOnTop; }
    void SetAlwaysOnTop( bool value ) { alwaysOnTop = value; if (parent) parent->BringToTop(this); }
    
    void SetMovable( bool value ) { movable = value; }

	void SetToolTip(const char* text) {toolTip.Replace(text);}

    /** Called whenever an item in a child list box is selected.
     *  @param selected  The listbox that had the item selected. 
     */
    virtual void OnListAction( pawsListBox* selected, int status ) {}
    
    /** Saves the position of this widget to the planeshift.cfg file.
     * This is done if the savedWindowPos is set to true, this can be done
     * inside the xml file as  savepositions="yes"
     */
    virtual void SavePosition();

    /** Get the position of this widget that was stored in a cfg file.
     * return A csrect that is the new position. Or return the position that is 
     * in the XML file.
     */
    virtual csRect LoadPositions();

    void SetFactory( const char* myfactory ) { factory = myfactory; }
    const char* GetType() { return factory; }

    int GetActualWidth( int myValue )
    {
        float desktop = (float)graphics2D->GetWidth();
        float value = (float)myValue;

        float ret = ( value * desktop ) / 800.0f + 0.5;
        return ( (int)ret );
    }

    int GetActualHeight( int myValue )
    {
        float desktop = (float)graphics2D->GetHeight();
        float value = (float)myValue;

        float ret = ( value * desktop ) / 600.0f + 0.5;
        return ( (int)ret );
    }

    virtual void MouseOver(bool value);

    void SetFade(bool value) { fade = value; }

    void DrawWidgetText(const char *text, int x, int y, int style=-1);

    /**
     * set the min height and width to protect the widget from negatives.
     **/
    void SetMinSize( int width, int height)
    {
        min_width = width;
        min_height = height;
    }

    /**
     * set the max size for height and width, default is screen height and width
     **/
    void SetMaxSize( int width, int height)
    {
        max_width = width;
        max_height = height;
    }

    /**
     * Sets name of the PAWS xml-file that describes context menu of our widget
     */
    void SetContextMenu(const csString & fileName) { contextMenuFile = fileName; }
   
    /**
     * If some part of the widget is not within the rectangle of its parent, then
     * the widget is moved inside (if it is small enough).
     */
     void MakeFullyVisible();
     
     /** This re-calculates a widget's on screen position to draw based on it's
       * relative position and parent screen location.
       */
     void RecalcScreenPositions();
     
     void ClipToParent();
     csRect ClipRect() { return clipRect; }

    /// Set Text Color
    void SetColour( int newColour = -1 );
    /// Set Font to use programmatically
    void SetFont( const char* fontName, int Size=0 );

    void ChangeFontSize( int newSize );
    int  GetFontSize();
    iFont *GetFont();
    int    GetFontColour();
    int    GetFontShadowColour();
    int    GetFontStyle();
    bool Contains( int x, int y );


    /** Creates a popup window that the user can use to
     *  adjust the settings of this window
     */
    virtual void CreateWidgetConfigWindow();
    
    /** Gets the minimim alpha value of this widget
     */
    int GetMinAlpha(){return alphaMin;};
    
    /** Gets the maximum alpha value of this widget
     */
    int GetMaxAlpha(){return alpha;};
    
    /** Gets the fade value of this widget
     */
    int GetFadeVal(){return (int)fadeVal;};

    /** Gets the minimim alpha of this widget
     */
    void SetMinAlpha(int value){alphaMin=value;};
    
    /** Gets the maximum alpha of this widget
     */
    void SetMaxAlpha(int value){alpha=value;};
    
    /** Gets the fade status of this widget
     */
    bool isFadeEnabled(){return fade;};
    
    /** Gets the fading speed
     */
    float GetFadeSpeed(){return fadeSpeed;};
    
    /** Sets the fading speed
     */
    void SetFadeSpeed(float speed){fadeSpeed=speed;};
    
    /** Sets the title of this widget (if it has a title bar)
     */
    void SetTitle(const char *title);
    
    /** Saves current widget settings to XML file  
     */
    virtual void SaveSettings();
    
    /** Loads current widget settings from XML file
     */
    virtual void LoadSettings();
    
    virtual void DrawToolTip(int x, int y);
    
    /** 
     */
    virtual void Dump(csString tab = "");
    
    /** Returns textual description of path to our widget through the widget tree 
      * (i.e. its direct and indirect parents)
      */
    csString GetPathInWidgetTree();

protected:
    /** Sets up the title bar for the widget.
     * 
     *  @param text The name that should go in the title.
     *  @param image The name of the backround image for the title
     *  @param align The alignment of the text
     *  @param close_button "yes" means to add a close button to the title bar.
     *
     *  @return true if all was successful.
     */
    bool SetTitle( const char* text, const char* image, const char* align, const char* close_button );
    
    /** This will check to see if the mouse is over the resize hot spot.
     */
    int ResizeFlags(int mouseX, int mouseY);
    /**
     * Creates context menu from file ('contextMenuFile' attribute)
     */
    bool CreateContextMenu();

    /**
     * Calculates the right position in array of children for a given child (depends on always-on-top state).
     */    
    int CalcChildPosition(pawsWidget * child);
    /**
     * Loads custom border color preferences.
     */
    void LoadBorderColours( iDocumentNode* node );
    
    void SetCloseButtonPos();
    
    /// factory name
    csString factory;

    /// An id for the widget. 
    int id;

    /// Convert from string flag to int flag.
    int GetAttachFlag( const char* flag );

    /// This widget's parent.
    pawsWidget* parent;

    csRef<iGraphics2D> graphics2D;
    
    PawsManager* windowManager;

    /// These are the screen coordinates that this widget should be drawn at.
    csRect screenFrame;
    
    csRect clipRect;

    /// The child widgets.
    /// Widgets are drawn from the end of this array to the beginning.
    /// Children with alwaysOnTop=true are positioned before all other children in this array.
    csArray<pawsWidget*> children;

    bool visible;

    /// Decides that this widget should write/load it's position from a config file.
    bool saveWidgetPositions;

	/// Decides if the settings (alpha, fade, etc) of this widget should be 
	/// configurable
	bool configurable;
	
	/// Decides if this widget is movable.
    bool movable;

    /// Decides if this widget can be resized by user. 
    bool isResizable;
    
    /// Is this windows painted on top of other windows (of the same parent) ?
    bool alwaysOnTop;

    /// Defines the minimum and maximum sizes that the widget can be
    int min_width;
    int min_height;
    int max_width;
    int max_height;

    /// The name of this widget.
    csString name;
    csString closeName;

    /// The default frame. Often times this is what is loaded from xml.
    csRect defaultFrame;

    /// Stores the bgColour for this widget.  -1 if no colour should be used.
    int bgColour;   

    /// Background image
    pawsImage* bgImage;
    
    pawsBorder* border;

    /// The attachpoints for when a resize takes place.
    int attachFlags;

    bool hasFocus;

	bool hasMouseFocus;

    /// The direction and percentage of current fade in progress.
    float fadeVal;

    /// The original background alpha.
    int alpha;

    /// The minimum background alpha.
    int alphaMin;

    /// Set to false to disable fading
    bool fade;

	/// The speed of the fading process
	float fadeSpeed;
    
    //// Path to the file that describes context menu of our witget (invoked by mouse right-click)
    //// If blank, then there is not context menu for our widget.
    csString contextMenuFile;
    
    //// Existing context menu of widget - NULL if there is no menu created at the moment.
    pawsMenu * contextMenu;

    /// Custom border colors for this window
    int borderColours[5];
    /// Flag to indicate using custom border colors
    bool hasBorderColours;

    /// optional font to use when drawing this widget
    csRef<iFont> myFont;
    /// optional color to use when drawing text with this widget
    int defaultFontColour;
    /// optional color to use when doing text dropshadows
    int defaultFontShadowColour;
    /// current font size
    int fontSize;
    /// name of currently selected font
    csString fontName;
    /// current font style
    int fontStyle;

    /// Used in the SelfPopulate functions to map to xml nodes for
    /// each widget
    csString xmlbinding;

	/// Tooltip to be displayed
	csString toolTip;
};

//------------------------------------------------------------



class pawsWidgetFactory
{
public:
    pawsWidgetFactory( PawsManager* manager );
    virtual ~pawsWidgetFactory();

    virtual pawsWidget* Create();

    virtual const char* GetName() { return factoryName; }

protected:
    virtual void Register( const char* name );

    csString factoryName;
    PawsManager* mainManager;
};

//---------------------------------------------------------------------------


class pawsBaseWidgetFactory : public pawsWidgetFactory
{
public:
    pawsBaseWidgetFactory( PawsManager* manager ):
        pawsWidgetFactory( manager )
    {
        Register( "pawsWidget" );
    }

    pawsWidget* Create()
    {
        return new pawsWidget( mainManager );
    }
};


#define CREATE_PAWS_FACTORY( factoryName ) \
class factoryName##Factory : public pawsWidgetFactory \
{ \
public: \
    factoryName##Factory( PawsManager* manager ) : pawsWidgetFactory( manager ) \
    { \
        Register( #factoryName ); \
    } \
 \
    pawsWidget* Create() \
    { \
        return new factoryName( mainManager ); \
    } \
};




#endif 


